"""
======================================
Reading NEXRAD Data from the AWS Cloud
======================================

Within this example, we show how you can remotely access Next Generation Weather Radar (NEXRAD) Data from Amazon Web Services
and plot quick looks of the datasets.

"""
print(__doc__)

# Author: Max Grover (mgrover@anl.gov)
# License: BSD 3 clause

import cartopy.crs as ccrs
import matplotlib.pyplot as plt

import pyart

######################################
# Read NEXRAD Level 2 Data
# ------------------------
#
# Let's start first with NEXRAD Level 2 data, which is ground-based radar data collected
# by the National Oceanic and Atmospheric Administration (NOAA), as a part of the National Weather Service
# ### Configure our Filepath for NEXRAD Level 2 Data
# We will access data from the **noaa-nexrad-level2** bucket, with the data organized as:
#
#    ``s3://noaa-nexrad-level2/year/month/date/radarsite/{radarsite}{year}{month}{date}_{hour}{minute}{second}_V06``
#
# Where in our case, we are using a sample data file from Houston, Texas (KHGX)
# on March 22, 2022, at 1201:25 UTC. This means our path would look like:

aws_nexrad_level2_file = (
    "s3://noaa-nexrad-level2/2022/03/22/KHGX/KHGX20220322_120125_V06"
)

#######################################
# We can use the **pyart.io.read_nexrad_archive** module to access our data, passing in the filepath.

radar = pyart.io.read_nexrad_archive(aws_nexrad_level2_file)

#######################################
# Let's take a look at a summary of what fields are available.

list(radar.fields)

#######################################
# Let's plot the reflectivity/velocity fields as a first step to investigating our dataset.
#
# Note: the reflectivity and velocity fields are in different sweeps, so we will need to specify which sweep to plot in each plot.

fig = plt.figure(figsize=(12, 4))
display = pyart.graph.RadarMapDisplay(radar)

ax = plt.subplot(121, projection=ccrs.PlateCarree())

display.plot_ppi_map(
    "reflectivity",
    sweep=0,
    ax=ax,
    colorbar_label="Equivalent Relectivity ($Z_{e}$) \n (dBZ)",
    vmin=-20,
    vmax=60,
)

ax = plt.subplot(122, projection=ccrs.PlateCarree())

display.plot_ppi_map(
    "velocity",
    sweep=1,
    ax=ax,
    colorbar_label="Radial Velocity ($V_{r}$) \n (m/s)",
    vmin=-70,
    vmax=70,
)

#######################################
# Within this plot, we see that the velocity data still has regions that are folded, indicating the dataset has not yet been dealiased.

#######################################
# Read NEXRAD Level 3 Data
# ------------------------
# We can also access NEXRAD Level 3 data using Py-ART!
#
# These datasets have had additional data quality processes applied, including dealiasing.
#
# Each Level 3 data field is stored in **separate file** - in this example, we will look at the reflectivity and
# velocity field at the lowest levels. These correspond to the following variable names:
#
# - ``N0U`` - Velocity at the lowest level
# - ``NOQ`` - Reflectivity at the lowest level
#
# These datasets are also in a different bucket (**unidata-nexrad-level3**), and the files are in a **flat directory structure**
# using the following naming convention:
#
#    ``s3://unidata-nexrad-level3/{radarsite}_{field}_{year}_{month}_{date}_{hour}_{minute}_{second}``
#
# For example, we can look at data from that same time as the NEXRAD Level 2 data used previously (March 22, 2022 at 1201 UTC)

aws_nexrad_level3_velocity_file = (
    "s3://unidata-nexrad-level3/HGX_N0U_2022_03_22_12_01_25"
)
aws_nexrad_level3_reflectivity_file = (
    "s3://unidata-nexrad-level3/HGX_N0Q_2022_03_22_12_01_25"
)

#######################################
# Read our Data using **pyart.io.read_nexrad_level3**

radar_level3_velocity = pyart.io.read_nexrad_level3(aws_nexrad_level3_velocity_file)
radar_level3_reflectivity = pyart.io.read_nexrad_level3(
    aws_nexrad_level3_reflectivity_file
)

#######################################
# Let's confirm that each radar object has a single field:

print(
    "velocity radar object: ",
    list(radar_level3_velocity.fields),
    "reflectivity radar object: ",
    list(radar_level3_reflectivity.fields),
)

#######################################
# Plot a Quick Look of our NEXRAD Level 3 Data
#
# Let's plot the reflectivity/velocity fields as a first step to investigating our dataset.
#
# Note: the reflectivity and velocity fields are in different radars, so we need to setup different displays.

fig = plt.figure(figsize=(12, 4))
reflectivity_display = pyart.graph.RadarMapDisplay(radar_level3_reflectivity)

ax = plt.subplot(121, projection=ccrs.PlateCarree())

reflectivity_display.plot_ppi_map(
    "reflectivity",
    ax=ax,
    colorbar_label="Equivalent Relectivity ($Z_{e}$) \n (dBZ)",
    vmin=-20,
    vmax=60,
)

velocity_display = pyart.graph.RadarMapDisplay(radar_level3_velocity)

ax = plt.subplot(122, projection=ccrs.PlateCarree())

velocity_display.plot_ppi_map(
    "velocity",
    ax=ax,
    colorbar_label="Radial Velocity ($V_{r}$) \n (m/s)",
    vmin=-70,
    vmax=70,
)



"""
Routines for reading IPHEx NOXP files.

"""

import datetime

import netCDF4
import numpy as np

from ..config import FileMetadata
from ..core.radar import Radar
from ..io.common import make_time_unit_str  # , _test_arguments

# Only providing updated names for the most common fields.
# Many more fields in the files, and these are read in but not renamed.
NOXP_FIELD_NAMES = {
    # uncorrected reflectivity, horizontal
    "ZH": "reflectivity",
    # uncorrected differential reflectivity
    "ZDR": "differential_reflectivity",
    "RHOHV": "cross_correlation_ratio",
    "KDP": "differential_phase",
    "KDPM": "specific_differential_phase",
    "VR": "velocity",
    "SVR": "spectrum_width",
}


def read_noxp_iphex_nc(
    filename,
    field_names=None,
    additional_metadata=None,
    file_field_names=False,
    exclude_fields=None,
    include_fields=None,
    **kwargs
):
    """
    Read a NOXP IPHEX netCDF file.

    Parameters
    ----------
    filename : str
        Name of the netCDF file to read.
    field_names : dict, optional
        Dictionary mapping netCDF field names to radar field names. If a
        data type found in the file does not appear in this dictionary or has
        a value of None it will not be placed in the radar.fields dictionary.
        A value of None, the default, will use the mapping defined in the
        Py-ART configuration file.
    additional_metadata : dict of dicts, optional
        Dictionary of dictionaries to retrieve metadata from during this read.
        This metadata is not used during any successive file reads unless
        explicitly included.  A value of None, the default, will not
        introduct any addition metadata and the file specific or default
        metadata as specified by the Py-ART configuration file will be used.
    file_field_names : bool, optional
        True to use the netCDF data type names for the field names. If this
        case the field_names parameter is ignored. The field dictionary will
        likely only have a 'data' key, unless the fields are defined in
        `additional_metadata`.
    exclude_fields : list or None, optional
        List of fields to exclude from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields specified by include_fields.
    include_fields : list or None, optional
        List of fields to include from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields not specified by exclude_fields.

    Returns
    -------
    radar : Radar
        Radar object containing data from netCDF file.

    """
    # TODO before moving to pyart.io
    # * unit test
    # * add default field mapping, etc to default config
    # * auto-detect file type with pyart.io.read function
    # * instrument parameters
    # * add additional checks for HOW attributes
    # * support for other objects (SCAN, XSEC)

    # test for non empty kwargs
    # Getting import error on this function, skipping for now.
    # Only issues a warning anyway.
    #    _test_arguments(kwargs)

    # create metadata retrieval object
    if field_names is None:
        field_names = NOXP_FIELD_NAMES
    filemetadata = FileMetadata(
        "cfradial",
        field_names,
        additional_metadata,
        file_field_names,
        exclude_fields,
        include_fields,
    )

    # read the data
    ncobj = netCDF4.Dataset(filename)
    ncvars = ncobj.variables

    # One sweep per file
    nsweeps = 1

    # latitude, longitude and altitude
    latitude = filemetadata("latitude")
    longitude = filemetadata("longitude")
    altitude = filemetadata("altitude")
    latitude["data"] = np.array(ncvars["Latitude"])
    longitude["data"] = np.array(ncvars["Longitude"])
    altitude["data"] = np.array(ncvars["Altitude"])

    # metadata
    metadata = filemetadata("metadata")
    metadata["source"] = "NOAA NSSL"
    metadata["original_container"] = "noxp_iphex_nc"
    metadata["system"] = ncobj.Radar

    # sweep_start_ray_index, sweep_end_ray_index
    sweep_start_ray_index = filemetadata("sweep_start_ray_index")
    sweep_end_ray_index = filemetadata("sweep_end_ray_index")
    rays_per_sweep = np.shape(ncvars["AZ"][:])
    ssri = np.cumsum(np.append([0], rays_per_sweep[:-1])).astype("int32")
    seri = np.cumsum(rays_per_sweep).astype("int32") - 1
    sweep_start_ray_index["data"] = ssri
    sweep_end_ray_index["data"] = seri

    # sweep_number
    sweep_number = filemetadata("sweep_number")
    sweep_number["data"] = np.arange(nsweeps, dtype="int32")

    # sweep_mode
    sweep_mode = filemetadata("sweep_mode")
    scan_type = getattr(ncobj, "Scan type").lower()
    if scan_type in ["ppi", "rhi"]:
        sweep_mode["data"] = np.array(nsweeps * ["manual_" + scan_type])
    else:  # Guessing that if not RHI or PPI, then a pointing scan
        sweep_mode["data"] = np.array(nsweeps * ["pointing"])

    # fixed_angle
    fixed_angle = filemetadata("fixed_angle")
    if scan_type == "rhi":
        sweep_el = ncvars["AZ"][0]
    else:
        sweep_el = ncvars["EL"][0]
    fixed_angle["data"] = np.array([sweep_el], dtype="float32")

    # elevation
    elevation = filemetadata("elevation")
    elevation["data"] = np.array(ncvars["EL"])

    # range
    _range = filemetadata("range")
    rng = 1000.0 * ncvars["Range"][:].T
    _range["data"] = rng[0]
    _range["meters_to_center_of_first_gate"] = np.mean(rng[0][0:2])
    # Gate spacing mostly but not completely constant
    _range["meters_between_gates"] = np.median(np.diff(_range["data"]))
    _range["spacing_is_constant"] = 0

    # azimuth
    azimuth = filemetadata("azimuth")
    azimuth["data"] = np.array(ncvars["AZ"])

    # time
    # Basing everything around 6/3/2014 because better than 1/0/0000
    # for datetime module.
    _time = filemetadata("time")
    frac_since_basetime = ncvars["time"][0] - 735762.0
    start_time = datetime.datetime(2014, 6, 3) + datetime.timedelta(frac_since_basetime)
    _time["units"] = make_time_unit_str(start_time)
    _time["data"] = (
        3600.0 * 24.0 * (ncvars["time"][:] - ncvars["time"][0]).astype("float32")
    )

    # fields
    # nearly all variables w/ dimensions of 'Gate', 'Time' are fields
    keys = [
        k
        for k, v in ncvars.items()
        if v.dimensions == ("Gate", "Time") and k not in ["Range", "Distance", "Height"]
    ]
    fields = {}
    for key in keys:
        field_name = filemetadata.get_field_name(key)
        if field_name is None:
            if exclude_fields is not None and key in exclude_fields:
                continue
            if include_fields is not None and key not in include_fields:
                continue
            field_name = key
        fields[field_name] = _ncvar_to_dict(ncvars[key])

    # instrument_parameters
    instrument_parameters = {}
    prt = filemetadata("prt")
    prt["data"] = 1.0 / float(ncobj.PRF[0:5])
    instrument_parameters["prt"] = prt
    frequency = filemetadata("frequency")
    frequency["frequency"] = float(ncobj.Frequency[0:4]) * 10e9
    instrument_parameters["frequency"] = frequency
    hits = filemetadata("n_samples")
    hits["data"] = ncobj.Hits
    instrument_parameters["hits"] = hits

    # Go get the radar object!
    return Radar(
        _time,
        _range,
        fields,
        metadata,
        scan_type,
        latitude,
        longitude,
        altitude,
        sweep_number,
        sweep_mode,
        fixed_angle,
        sweep_start_ray_index,
        sweep_end_ray_index,
        azimuth,
        elevation,
        instrument_parameters=instrument_parameters,
    )


def _ncvar_to_dict(ncvar):
    """Convert a NetCDF Dataset variable to a dictionary."""
    # copy all attribute except for scaling parameters
    d = {
        k: getattr(ncvar, k)
        for k in ncvar.ncattrs()
        if k not in ["scale_factor", "add_offset"]
    }
    d["data"] = ncvar[:].T  # Data originally stored as (gate, azimuth)
    if np.isscalar(d["data"]):
        # netCDF4 1.1.0+ returns a scalar for 0-dim array, we always want
        # 1-dim+ arrays with a valid shape.
        d["data"] = np.array(d["data"])
        d["data"].shape = (1,)
    return d



"""
================================
Correct reflectivity attenuation
================================

In this example the reflectivity attenuation is calculated and then corrected
for a polarimetric radar using a Z-PHI method implemented in Py-ART.

"""
print(__doc__)

# Author: Jonathan J. Helmus (jhelmus@anl.gov)
# License: BSD 3 clause

import matplotlib.pyplot as plt

import pyart

file = pyart.testing.get_test_data("sgpcsaprsurcmacI7.c0.20110520.095101.nc")

# read in the data
radar = pyart.io.read_cfradial(file)

# remove existing corrections
radar.fields.pop("specific_attenuation")
radar.fields.pop("corrected_reflectivity_horizontal")

# perform attenuation correction
spec_at, cor_z = pyart.correct.calculate_attenuation(
    radar,
    0,
    refl_field="reflectivity_horizontal",
    ncp_field="norm_coherent_power",
    rhv_field="copol_coeff",
    phidp_field="proc_dp_phase_shift",
)
radar.add_field("specific_attenuation", spec_at)
radar.add_field("corrected_reflectivity_horizontal", cor_z)

# create the plot
fig = plt.figure(figsize=(15, 5))
ax1 = fig.add_subplot(131)
display = pyart.graph.RadarDisplay(radar)
display.plot(
    "reflectivity_horizontal",
    0,
    ax=ax1,
    vmin=0,
    vmax=60.0,
    colorbar_label="",
    title="Raw Reflectivity",
)

ax2 = fig.add_subplot(132)
display.plot(
    "specific_attenuation",
    0,
    vmin=0,
    vmax=1.0,
    colorbar_label="",
    ax=ax2,
    title="Specific Attenuation",
)

ax3 = fig.add_subplot(133)
display = pyart.graph.RadarDisplay(radar)
display.plot(
    "corrected_reflectivity_horizontal",
    0,
    vmin=0,
    vmax=60.0,
    colorbar_label="",
    ax=ax3,
    title="Corrected Reflectivity",
)

plt.suptitle("Attenuation correction using Py-ART", fontsize=16)
plt.show()



"""
=================================
Modify a Colorbar for your Plot
=================================

This is an example of how to modify a colobar
within a Py-ART display object.

"""
print(__doc__)

# Author: Joe O'Brien (obrienj@anl.gov)
# License: BSD 3 clause

import cartopy.crs as ccrs
import matplotlib.pyplot as plt

import pyart
from pyart.testing import get_test_data

#########################################
# ** Initial PPI Map Display
#
# Let's plot a PPI Map Display
# and take a look at the colorbar
#
# Notice: the colorbar is not perfect
# and slightly overlaps the PPI display

# Define figure
fig = plt.figure()

# Create a subplot with correct cartopy projection
axs = plt.subplot(111, projection=ccrs.PlateCarree())

# Define and Read in the test data
radar_file = get_test_data("swx_20120520_0641.nc")
radar = pyart.io.read(radar_file)

# Create the Radar Map Display (defines x,y as lat/lons)
display = pyart.graph.RadarMapDisplay(radar)

# Display the horizontal equivalent reflectivity factor
# Note: embellish = False will remove initial lat/lon coords
display.plot_ppi_map(
    "reflectivity_horizontal",
    2,
    ax=axs,
    vmin=-30,
    vmax=60,
    embellish=False,
    norm=None,
    cmap="pyart_HomeyerRainbow",
)

# Add gridlines
gl = axs.gridlines(
    crs=ccrs.PlateCarree(),
    draw_labels=True,
    linewidth=1,
    color="gray",
    alpha=0.3,
    linestyle="--",
)

plt.gca().xaxis.set_major_locator(plt.NullLocator())

# Make sure labels are only plotted on the left and bottom
gl.top_labels = False
gl.right_labels = False

gl.xlabel_style = {"size": 14}
gl.ylabel_style = {"size": 14}

# delete the display object
del display

############################################
# ** Colorbar Position / Title Manipulation

# Now, let's update the colorbar position
# to match the display

# Define figure
fig = plt.figure()

# Create a subplot with correct cartopy projection
axsB = plt.subplot(111, projection=ccrs.PlateCarree())

# Create the Radar Map Display (defines x,y as lat/lons)
display = pyart.graph.RadarMapDisplay(radar)

# Create the display again
ppi_map = display.plot_ppi_map(
    "reflectivity_horizontal",
    2,
    ax=axsB,
    vmin=-30,
    vmax=60,
    embellish=False,
    norm=None,
    cmap="pyart_HomeyerRainbow",
)

# Add gridlines
gl = axsB.gridlines(
    crs=ccrs.PlateCarree(),
    draw_labels=True,
    linewidth=1,
    color="gray",
    alpha=0.3,
    linestyle="--",
)

plt.gca().xaxis.set_major_locator(plt.NullLocator())

# Make sure labels are only plotted on the left and bottom
gl.top_labels = False
gl.right_labels = False

# Define the size of the grid labels
gl.xlabel_style = {"size": 12}
gl.ylabel_style = {"size": 12}

# Define the colorbar from the RadarMapDisplay object
cbar = display.cbs[0]
# Modify the colorbar label and size
cbar.set_label(label="Horizontal Reflectivity Factor ($Z_{H}$) (dBZ)", fontsize=12)
# Modify the number of colorbar ticks
cbar.set_ticks([-20, 0, 20, 40, 60])



#! /usr/bin/env python
"""
Make a small MDV file containing a single RHI scan.

Single field and scan is taken from full sized file 110041.mdv
"""

import gzip
import struct

import StringIO

import pyart

# MDV file parameters
MASTER_HEADER_SIZE = 1024  # size of master header in bytes
FIELD_HEADER_SIZE = 416  # size of field header in bytes
VLEVEL_HEADER_SIZE = 1024  # size of vlevel header in bytes
CHUNK_HEADER_SIZE = 512  # size of chunk header in bytes
COMPRESSION_INFO_SIZE = 24  # size of compression infomation header in bytes
SWEEP_INFO_SIZE = 8  # size of sweep info header in bytes

# parameters
FIELD_NUMBER = 3  # field to extract in new file (3 is reflectivity)
NGATES = 125
NFIELDS = 1
INFILE = "110041.mdv"
OUTFILE = "example_mdv_rhi.mdv"

# read and compress the reflectivity field data
mdvfile = pyart.io.mdv.MdvFile(INFILE)
number_of_fields = int(mdvfile.master_header["nfields"])
bias = mdvfile.field_headers[FIELD_NUMBER]["bias"]
scale = mdvfile.field_headers[FIELD_NUMBER]["scale"]
in_field_offset = mdvfile.field_headers[FIELD_NUMBER]["field_data_offset"]

# bais, scale and slice to extract the first NGATES
fdata = ((mdvfile.read_a_field(FIELD_NUMBER) - bias) / scale)[0, :, :NGATES]
mdvfile.close()

fdata_str = fdata.astype("uint16").byteswap().tostring()
uncompressed_data_size = len(fdata_str)
fileobj = StringIO.StringIO()
gzipfile = gzip.GzipFile(fileobj=fileobj, mode="w")
gzipfile.write(fdata_str)
gzipfile.close()
compressed_field_data = fileobj.getvalue()
compressed_data_size = len(compressed_field_data)

# prepare input and output files for reading/writing
f = open(INFILE, "rb")
out = open(OUTFILE, "wb")

# read the master header, update, and write
fmt = ">28i 8i i 5i 6f 3f 12f 512c 128c 128c i"
packet = list(struct.unpack(fmt, f.read(struct.calcsize(fmt))))
packet[19] = NFIELDS  # nfields
packet[20] = NGATES  # ngates
packet[24] = MASTER_HEADER_SIZE  # field_hdr_offset
packet[25] = MASTER_HEADER_SIZE + FIELD_HEADER_SIZE  # vlevel_hdr_offset
# chunk_hdr _offset
packet[26] = MASTER_HEADER_SIZE + FIELD_HEADER_SIZE + VLEVEL_HEADER_SIZE
out.write(struct.pack(fmt, *packet))

# read the reflectivity field header, update and write
f.seek(MASTER_HEADER_SIZE + FIELD_NUMBER * FIELD_HEADER_SIZE)
fmt = ">17i 10i 9i 4i f f 8f 12f 4f 5f 64c 16c 16c 16c 16c i"
packet = list(struct.unpack(fmt, f.read(struct.calcsize(fmt))))
packet[9] = NGATES  # ngates
field_data_offset = (
    MASTER_HEADER_SIZE
    + FIELD_HEADER_SIZE
    + VLEVEL_HEADER_SIZE
    + CHUNK_HEADER_SIZE * FIELD_NUMBER
)
packet[15] = field_data_offset  # field_data_offset
# volume_size
packet[16] = compressed_data_size + COMPRESSION_INFO_SIZE + SWEEP_INFO_SIZE
out.write(struct.pack(fmt, *packet))

# read the reflectivity vlevel header, update and write
f.seek(
    MASTER_HEADER_SIZE
    + FIELD_HEADER_SIZE * number_of_fields
    + FIELD_NUMBER * VLEVEL_HEADER_SIZE
)
fmt = ">i i 122i 4i 122f 5f i"
packet = list(struct.unpack(fmt, f.read(struct.calcsize(fmt))))
out.write(struct.pack(fmt, *packet))

# read the three chunk headers, update and write
f.seek(
    MASTER_HEADER_SIZE
    + FIELD_HEADER_SIZE * number_of_fields
    + VLEVEL_HEADER_SIZE * number_of_fields
)
fmt = ">5i 2i 480c i"

chunk_data_offset = (
    compressed_data_size + field_data_offset + SWEEP_INFO_SIZE + COMPRESSION_INFO_SIZE
)

# radar_info chunk header
packet = list(struct.unpack(fmt, f.read(struct.calcsize(fmt))))
info_chunk_offset = int(packet[3])
packet[3] = chunk_data_offset
out.write(struct.pack(fmt, *packet))

# calib chunk header
packet = list(struct.unpack(fmt, f.read(struct.calcsize(fmt))))
calib_chunk_offset = int(packet[3])
packet[3] = chunk_data_offset + 240
out.write(struct.pack(fmt, *packet))

# elevs chunk header
packet = list(struct.unpack(fmt, f.read(struct.calcsize(fmt))))
elevs_chunk_offset = int(packet[3])
packet[3] = chunk_data_offset + 240 + 300
out.write(struct.pack(fmt, *packet))

# read the sweep_info and write
f.seek(in_field_offset)
fmt = "I I"
packet = list(struct.unpack(fmt, f.read(struct.calcsize(fmt))))
out.write(struct.pack(fmt, *packet))

# read the compression header, update and write
fmt = ">I I I I 2I"
packet = list(struct.unpack(fmt, f.read(struct.calcsize(fmt))))
packet[1] = uncompressed_data_size
packet[2] = compressed_data_size + COMPRESSION_INFO_SIZE  # nbytes_compressed
packet[3] = compressed_data_size  # nytes_coded
out.write(struct.pack(fmt, *packet))

# write the compressed field data
out.write(compressed_field_data)

# read/write radar_info chunk
f.seek(info_chunk_offset)
fmt = ">12i 2i 22f 4f 40c 40c"
packet = list(struct.unpack(fmt, f.read(struct.calcsize(fmt))))
packet[2] = NFIELDS  # nfield
packet[3] = NGATES  # ngates
out.write(struct.pack(fmt, *packet))

# read/write calib chunk
f.seek(calib_chunk_offset)
fmt = ">16c 6i 51f 14f"
packet = list(struct.unpack(fmt, f.read(struct.calcsize(fmt))))
out.write(struct.pack(fmt, *packet))

# read/write elevs chunk
f.seek(elevs_chunk_offset)
fmt = "%df" % (2)
packet = list(struct.unpack(fmt, f.read(struct.calcsize(fmt))))
out.write(struct.pack(fmt, *packet))

# close the output file
out.close()



""" Unit Tests for Py-ART's io/uf.py and io/uffile.py modules. """

from datetime import datetime

try:
    from StringIO import StringIO
except ImportError:
    from io import BytesIO as StringIO

import numpy as np
import pytest
from numpy.testing import assert_almost_equal, assert_warns

import pyart
from pyart.io.uffile import UFFile, UFRay

radar = pyart.io.read_uf(pyart.testing.UF_FILE, file_field_names=True)


def test_time():
    assert radar.time["units"] == "seconds since 2011-05-20T10:54:16Z"
    assert_almost_equal(radar.time["data"][0], 0)


def test_range():
    assert np.allclose(radar.range["data"], np.arange(667) * 60 + 30)
    assert radar.range["meters_to_center_of_first_gate"] == 30
    assert radar.range["meters_between_gates"] == 60


def test_lat_lon_alt():
    assert_almost_equal(radar.latitude["data"], 36.49, 2)
    assert_almost_equal(radar.longitude["data"], -97.59, 2)
    assert_almost_equal(radar.altitude["data"], 214)


def test_sweep_start_ray_index():
    assert np.allclose(radar.sweep_start_ray_index["data"], np.array([0]))


def test_sweep_end_ray_index():
    assert np.allclose(radar.sweep_end_ray_index["data"], np.array([0]))


def test_sweep_number():
    assert np.allclose(radar.sweep_number["data"], np.array([0]))


def test_scan_type_ppi():
    assert radar.scan_type == "ppi"


def test_sweep_mode():
    assert radar.sweep_mode["data"].dtype.char == "S"
    assert radar.sweep_mode["data"] == np.array([b"azimuth_surveillance"])


def test_fixed_angle():
    assert np.allclose(radar.fixed_angle["data"], np.array([0.5]))


def test_elevation():
    assert np.allclose(radar.elevation["data"], np.array([0.484375]))


def test_azimuth():
    assert np.allclose(radar.azimuth["data"], np.array([359.9375]))


fields = {
    "CZ": 0.0,
    "DR": 0.0,
    "DZ": -6.0499999999999998,
    "HC": 1.0,
    "KD": 0.0,
    "PH": 90.0,
    "RH": 1.0,
    "SQ": 1.0,
    "SW": 0.01,
    "VR": 0.0,
    "ZD": 0.0,
    "ZT": -6.0499999999999998,
}


@pytest.mark.parametrize("field, field_value", fields.items(), ids=list(fields.keys()))
def test_fields(field, field_value):
    check_field(field, field_value)


def check_field(field, value):
    assert_almost_equal(radar.fields[field]["data"][0, 0], value)


def test_raises_ioerror():
    fake_bad_file = StringIO(b"XXXXXXXX")
    pytest.raises(IOError, pyart.io.read_uf, fake_bad_file)


def test_read_fileobj():
    fh = open(pyart.testing.UF_FILE, "rb")
    pyart.io.read_uf(fh)
    fh.close()


def test_instrument_parameters():
    # test that instrument parameters exist
    assert radar.scan_rate is not None
    assert "pulse_width" in radar.instrument_parameters
    assert "radar_beam_width_h" in radar.instrument_parameters
    assert "radar_beam_width_v" in radar.instrument_parameters
    assert "radar_receiver_bandwidth" in radar.instrument_parameters
    assert "polarization_mode" in radar.instrument_parameters
    assert "frequency" in radar.instrument_parameters
    assert "prt" in radar.instrument_parameters
    assert "nyquist_velocity" in radar.instrument_parameters


def test_nyquist_vel():
    ufile = UFFile(pyart.testing.UF_FILE)
    ufile.rays[0].field_headers[1].pop("nyquist")
    assert ufile.get_nyquists() is None

    ufile = UFFile(pyart.testing.UF_FILE)
    ray = UFRay(ufile.rays[0]._buf)
    ray.field_headers[1].pop("nyquist")
    ufile.rays.append(ray)
    ufile.nrays = 2
    assert ufile.get_nyquists() is None


def test_datetime():
    ufile = UFFile(pyart.testing.UF_FILE)
    ray = ufile.rays[0]
    assert ray.get_datetime() == datetime(2011, 5, 20, 10, 54, 16)

    # test case where midnight is incorrectly represented as 24:00:00
    ray.mandatory_header["hour"] = 24
    ray.mandatory_header["minute"] = 0
    ray.mandatory_header["second"] = 0
    assert ray.get_datetime() == datetime(2011, 5, 21, 0, 0, 0)


def test_polarization():
    ufile = UFFile(pyart.testing.UF_FILE)
    ufile.rays[0].field_headers[0]["polarization"] = 99
    assert ufile.get_sweep_polarizations()[0] == "elliptical"


def test_frequency():
    ufile = UFFile(pyart.testing.UF_FILE)
    filemetadata = pyart.config.FileMetadata("uf")

    ip = pyart.io.uf._get_instrument_parameters(ufile, filemetadata)
    assert "frequency" in ip
    assert_almost_equal(ip["frequency"]["data"][0], 9.69026150e09, -3)

    # An invalid wavelength should throw a warning
    ufile.rays[0].field_headers[0]["wavelength_cm"] = 0
    assert_warns(
        UserWarning, pyart.io.uf._get_instrument_parameters, ufile, filemetadata
    )


def test_scan_type_uf():
    ufile = UFFile(pyart.testing.UF_FILE)
    ufray = ufile.rays[0]

    scan_type = pyart.io.uf._get_scan_type(ufray)
    assert scan_type == "ppi"

    # An invalid scan mode should throw a warning
    ufray.mandatory_header["sweep_mode"] = 99
    assert_warns(UserWarning, pyart.io.uf._get_scan_type, ufray)


def test_skip_field():
    test_radar = pyart.io.read_uf(
        pyart.testing.UF_FILE, exclude_fields=["DZ"], file_field_names=True
    )
    assert "DZ" not in test_radar.fields.keys()


def test_range_start_nonzero():
    radar = pyart.io.read_uf(pyart.testing.UF_FILE)
    radar.range["meters_to_center_of_first_gate"] += 1500.0
    in_mem = StringIO()
    pyart.io.write_uf(in_mem, radar)

    in_mem.seek(0)
    radar2 = pyart.io.read_uf(in_mem)
    assert_almost_equal(radar2.range["meters_to_center_of_first_gate"], 1530)
    assert_almost_equal(radar2.range["data"][0], 1530)



"""
Routines for reading files from the X-band radar from the PATTERN_ project.

.. _PATTERN: http://www.mi.uni-hamburg.de/PATTERN-Pre.6763.0.html

"""

import datetime

import netCDF4
import numpy as np

from ..config import FileMetadata
from ..core.radar import Radar
from ..io.common import _test_arguments, make_time_unit_str


def read_pattern(filename, **kwargs):
    """
    Read a netCDF file from a PATTERN project X-band radar.

    Parameters
    ----------
    filename : str
        Name of netCDF file to read data from.

    Returns
    -------
    radar : Radar
        Radar object.

    """
    # test for non empty kwargs
    _test_arguments(kwargs)

    # create metadata retrieval object
    filemetadata = FileMetadata("pattern")

    # read the data
    ncobj = netCDF4.Dataset(filename)
    ncvars = ncobj.variables

    # general parameters
    nrays = ncvars["Azimuth"].shape[0]
    scan_type = "ppi"

    # time
    # interpolate between the first and last timestamps in the Time variable
    time = filemetadata("time")
    nctime = ncvars["Time"]
    time["units"] = make_time_unit_str(datetime.datetime.utcfromtimestamp(nctime[0]))
    time["data"] = np.linspace(0, nctime[-1] - nctime[0], nrays)

    # range
    _range = filemetadata("range")
    _range["data"] = ncvars["Distance"][:]
    _range["meters_to_center_of_first_gate"] = _range["data"][0]
    # assuming the distance between all gates is constant, may not
    # always be true.
    _range["meters_between_gates"] = _range["data"][1] - _range["data"][0]

    # fields
    # files contain a single corrected reflectivity field
    fields = {}
    field_name = filemetadata.get_field_name("corrected_reflectivity")
    field_dic = filemetadata(field_name)
    field_dic["_FillValue"] = ncvars["Corrected_Reflectivity"]._FillValue
    field_dic["data"] = ncvars["Corrected_Reflectivity"][:]
    fields[field_name] = field_dic

    # metadata
    metadata = filemetadata("metadata")
    for k in ["institution", "title", "used_algorithms"]:
        if k in ncobj.ncattrs():
            metadata[k] = ncobj.getncattr(k)

    # latitude, longitude, altitude
    latitude = filemetadata("latitude")
    longitude = filemetadata("longitude")
    altitude = filemetadata("altitude")
    latitude["data"] = np.array([ncobj.latitude[:-1]], dtype="float64")
    longitude["data"] = np.array([ncobj.longitude[:-1]], dtype="float64")
    altitude["data"] = np.array([ncobj.elevation], dtype="float64")

    # sweep parameters
    # sweep_number, sweep_mode, fixed_angle, sweep_start_ray_index,
    # sweep_end_ray_index
    sweep_number = filemetadata("sweep_number")
    sweep_mode = filemetadata("sweep_mode")
    fixed_angle = filemetadata("fixed_angle")
    sweep_start_ray_index = filemetadata("sweep_start_ray_index")
    sweep_end_ray_index = filemetadata("sweep_end_ray_index")

    sweep_number["data"] = np.arange(1, dtype="int32")
    sweep_mode["data"] = np.array(1 * ["azimuth_surveillance"])
    fixed_angle["data"] = np.array([0], dtype="float32")
    sweep_start_ray_index["data"] = np.array([0], dtype="int32")
    sweep_end_ray_index["data"] = np.array([nrays - 1], dtype="int32")

    # azimuth, elevation
    azimuth = filemetadata("azimuth")
    elevation = filemetadata("elevation")

    azimuth["data"] = ncvars["Azimuth"][:]
    elevation["data"] = np.array([0.0], dtype="float32")

    # instrument parameters
    instrument_parameters = None

    return Radar(
        time,
        _range,
        fields,
        metadata,
        scan_type,
        latitude,
        longitude,
        altitude,
        sweep_number,
        sweep_mode,
        fixed_angle,
        sweep_start_ray_index,
        sweep_end_ray_index,
        azimuth,
        elevation,
        instrument_parameters=instrument_parameters,
    )



""" Unit Tests for Py-ART's io/nexrad_cdm.py module. """

import bz2

import numpy as np
import pytest
from numpy.ma.core import MaskedArray
from numpy.testing import assert_almost_equal

import pyart

###################################################
# read_nexrad_cdm tests (verify radar attributes) #
###################################################

# read in the sample file and create the radar objects
# We need to decompress the bz2 file which contains this data
with pyart.testing.InTemporaryDirectory():
    tmpfile = "tmp_nexrad.nc"
    with open(tmpfile, "wb") as f:
        f.write(bz2.BZ2File(pyart.testing.NEXRAD_CDM_FILE).read())
    radar = pyart.io.read_nexrad_cdm(tmpfile)


# time attribute
def test_time():
    assert "comment" in radar.time.keys()
    assert "long_name" in radar.time.keys()
    assert "standard_name" in radar.time.keys()
    assert "units" in radar.time.keys()
    assert "calendar" in radar.time.keys()
    assert "data" in radar.time.keys()
    assert radar.time["units"] == "seconds since 2013-07-17T19:50:21Z"
    assert radar.time["data"].shape == (7200,)
    assert_almost_equal(radar.time["data"][1], 0.677, 3)


# range attribute
def test_range():
    assert "long_name" in radar.range
    assert "standard_name" in radar.range
    assert "meters_to_center_of_first_gate" in radar.range
    assert "meters_between_gates" in radar.range
    assert "units" in radar.range
    assert "data" in radar.range
    assert "spacing_is_constant" in radar.range
    assert radar.range["data"].shape == (1832,)
    assert_almost_equal(radar.range["data"][1], 2375, 0)


# fields attribute is tested later


# metadata attribute
def test_metadata():
    assert "instrument_name" in radar.metadata
    assert "source" in radar.metadata


# scan_type attribute
def test_scan_type():
    assert radar.scan_type == "ppi"


# latitude attribute
def test_latitude():
    assert "data" in radar.latitude
    assert "standard_name" in radar.latitude
    assert "units" in radar.latitude
    assert radar.latitude["data"].shape == (1,)
    assert_almost_equal(radar.latitude["data"], 48, 0)


# longitude attribute
def test_longitude():
    assert "data" in radar.longitude
    assert "standard_name" in radar.longitude
    assert "units" in radar.longitude
    assert radar.longitude["data"].shape == (1,)
    assert_almost_equal(radar.longitude["data"], -122, 0)


# altitude attribute
def test_altitude():
    assert "data" in radar.altitude
    assert "standard_name" in radar.altitude
    assert "units" in radar.altitude
    assert "positive" in radar.altitude
    assert radar.altitude["data"].shape == (1,)
    assert_almost_equal(radar.altitude["data"], 151, 0)  # 10 m different


# altitude_agl attribute
def test_altitude_agl():
    assert radar.altitude_agl is None


# sweep_number attribute
def test_sweep_number():
    assert "standard_name" in radar.sweep_number
    assert np.all(radar.sweep_number["data"] == range(16))


# sweep_mode attribute
def test_sweep_mode():
    assert "standard_name" in radar.sweep_mode
    assert radar.sweep_mode["data"].shape == (16,)
    assert radar.sweep_mode["data"].dtype.char == "S"
    assert np.all(radar.sweep_mode["data"] == [b"azimuth_surveillance"])


# fixed_angle attribute
def test_fixed_angle():
    assert "standard_name" in radar.fixed_angle
    assert "units" in radar.fixed_angle
    assert radar.fixed_angle["data"].shape == (16,)
    assert_almost_equal(radar.fixed_angle["data"][0], 0.53, 2)


# sweep_start_ray_index attribute
def test_sweep_start_ray_index():
    assert "long_name" in radar.sweep_start_ray_index
    assert radar.sweep_start_ray_index["data"].shape == (16,)
    assert_almost_equal(radar.sweep_start_ray_index["data"][0], 0, 0)


# sweep_end_ray_index attribute
def test_sweep_end_ray_index():
    assert "long_name" in radar.sweep_end_ray_index
    assert radar.sweep_end_ray_index["data"].shape == (16,)
    assert_almost_equal(radar.sweep_end_ray_index["data"][0], 719, 0)


# target_scan_rate attribute
def test_target_scan_rate():
    assert radar.target_scan_rate is None


# azimuth attribute
def test_azimuth():
    assert "standard_name" in radar.azimuth
    assert "long_name" in radar.azimuth
    assert "units" in radar.azimuth
    assert "axis" in radar.azimuth
    assert_almost_equal(radar.azimuth["data"][0], 350, 0)
    assert_almost_equal(radar.azimuth["data"][10], 355, 0)


# elevation attribute
def test_elevation():
    assert "standard_name" in radar.elevation
    assert "long_name" in radar.azimuth
    assert "units" in radar.elevation
    assert "axis" in radar.elevation
    assert radar.elevation["data"].shape == (7200,)
    assert_almost_equal(radar.elevation["data"][0], 0.75, 2)


# scan_rate attribute
def test_scan_rate():
    assert radar.scan_rate is None


# antenna_transition attribute
def test_antenna_transition():
    assert radar.antenna_transition is None


# instrument_parameters attribute
def test_instument_parameters():
    assert radar.instrument_parameters is None


# radar_calibration attribute
def test_radar_calibration():
    assert radar.radar_calibration is None


# ngates attribute
def test_ngates():
    assert radar.ngates == 1832


# nrays attribute
def test_nrays():
    assert radar.nrays == 7200


# nsweeps attribute
def test_nsweeps():
    assert radar.nsweeps == 16


####################
# fields attribute #
####################

fields = [
    "differential_phase",
    "spectrum_width",
    "cross_correlation_ratio",
    "reflectivity",
    "differential_reflectivity",
    "velocity",
]


@pytest.mark.parametrize("field", fields)
def test_field_dics(field):
    description = "field : %s, dictionary" % field
    check_field_dic.description = description
    check_field_dic(field)


def check_field_dic(field):
    """Check that the required keys are present in a field dictionary."""
    assert "standard_name" in radar.fields[field]
    assert "units" in radar.fields[field]
    assert "_FillValue" in radar.fields[field]
    assert "coordinates" in radar.fields[field]


@pytest.mark.parametrize("field", fields)
def test_field_shapes(field):
    description = "field : %s, shape" % field
    check_field_shape.description = description
    check_field_shape(field)


def check_field_shape(field):
    assert radar.fields[field]["data"].shape == (7200, 1832)


fields = {
    "differential_phase": MaskedArray,
    "spectrum_width": MaskedArray,
    "cross_correlation_ratio": MaskedArray,
    "reflectivity": MaskedArray,
    "differential_reflectivity": MaskedArray,
    "velocity": MaskedArray,
}


@pytest.mark.parametrize("field, field_type", fields.items(), ids=list(fields.keys()))
def test_field_types(field, field_type):
    description = "field : %s, type" % field
    check_field_type.description = description
    check_field_type(field, field_type)


def check_field_type(field, field_type):
    assert type(radar.fields[field]["data"]) is field_type


fields = {
    "differential_phase": 181.0,
    "spectrum_width": np.ma.masked,
    "cross_correlation_ratio": 0.0,
    "reflectivity": -32.0,
    "differential_reflectivity": -8.0,
    "velocity": np.ma.masked,
}


@pytest.mark.parametrize("field, field_value", fields.items(), ids=list(fields.keys()))
def test_field_first_points(field, field_value):
    # these values can be found using:
    # [round(radar.fields[f]['data'][0,0]) for f in radar.fields]
    description = "field : %s, first point" % field
    check_field_first_point.description = description
    check_field_first_point(field, field_value)


def check_field_first_point(field, value):
    if np.ma.is_masked(value):
        assert np.ma.is_masked(radar.fields[field]["data"][0, 0])
    else:
        assert_almost_equal(radar.fields[field]["data"][0, 0], value, 0)



#! /usr/bin/env python
import matplotlib.pyplot as plt

import pyart

# plot quickly
pradar = pyart.io.read_mdv("example_mdv_rhi.mdv")
display = pyart.graph.RadarDisplay(pradar)
fig = plt.figure()
ax = fig.add_subplot(111)
display.plot_rhi("reflectivity_horizontal", 0, vmin=-16, vmax=64)
fig.savefig("example_mdv_rhi.png")



"""
Configuration file for the Python ARM Radar Toolkit (Py-ART)

The values for a number of Py-ART parameters and the default metadata created
when reading files, correcting fields, etc. is controlled by this single
Python configuration file.

Py-ART's configuration can be modified by setting the environmental
variable PYART_CONFIG to point to a configuration file with formatting similar
to this file.

The recommended method for changing these defaults is for users to copy this
file into their home directory, rename it to .pyart_config.py, make any
desired changes, and adjust their login scripts to set the PYART_CONFIG
environmental variable to point to .pyart_config.py in their home directory.

Py-ART's configuration can also be modified within a script or shell session
using the load_config functions, such modification will last until a the end
of the script/session or until a new configuration is loaded.

"""

# THIS FILE WAS EDITED FOR USE IN PY-ART UNIT TESTS, changed noted by
# XXX changed

##############################################################################
##############################################################################
# Simple configuration
#
# Adjust the values of the variable (left hand side of the equal sign) in this
# section for an easy method of customizing Py-ART.  Do not change the
# variables names (the right hand side of the equal sign).  More advanced
# settings are based upon these variables.  Most users will find that
# adjusting this section is all that is needed.
##############################################################################
##############################################################################

# The default fill value for masked arrays and _FillValue keys
fill_value = -5555.0  # XXX changed

# Field names used when reading in radar files and in the various correction
# and retrieval algorithms. The comments in this section provide additional
# information about the fields in that section.

# radar reflectivity fields,
reflectivity = "REF"  # XXX changed
corrected_reflectivity = "corrected_reflectivity"
total_power = "total_power"

# mean doppler velocity fields, VEL
velocity = "velocity"
corrected_velocity = "corrected_velocity"

# spectral width field, SW
spectrum_width = "spectrum_width"

# differential reflectivity fields, ZDR
differential_reflectivity = "differential_reflectivity"
corrected_differential_reflectivity = "corrected_differential_reflectivity"

# cross correlation ratio, correlation coefficient, RhoHV
cross_correlation_ratio = "cross_correlation_ratio"

# normalized coherent power, signal quality index, SQI, NCP
normalized_coherent_power = "normalized_coherent_power"

# Differential phase shift, PhiDP
differential_phase = "differential_phase"
unfolded_differential_phase = "unfolded_differential_phase"
corrected_differential_phase = "corrected_differential_phase"

# Specific differential phase shift, KDP
specific_differential_phase = "specific_differential_phase"
corrected_specific_differential_phase = "corrected_specific_differential_phase"

# Linear depolarization ration (h - horizontal, v - vertical), LDR
linear_depolarization_ratio = "linear_polarization_ratio"
linear_depolarization_ratio_h = "linear_polarization_ratio_h"
linear_depolarization_ratio_v = "linear_polarization_ratio_v"

# Misc fields
signal_to_noise_ratio = "signal_to_noise_ratio"
rain_rate = "rain_rate"
radar_estimated_rain_rate = "radar_estimated_rain_rate"
radar_echo_classification = "radar_echo_classification"
specific_attenuation = "specific_attenuation"

# End of Simple Configuration section

##############################################################################
##############################################################################
# Advanced Configuration
#
# Most users will not want to make any changes in this section.  For users
# who want a more fined grained control over Py-ART's configuration this
# section provides access to these controls.  The layout of this section can
# be changed, the only requirement for a valid configuration file is that
# the ALL CAPITALIZED variable must must be present with the formatting
# present in this file.  These required variables are:
#
# FILL_VALUE, DEFAULT_METADATA, FILE_SPECIFIC_METADATA, FIELD_MAPPINGS,
# DEFAULT_FIELD_NAMES
#
# This section makes generous use of the variables in the Simple Configuration
# section, this is not required, but simplifies and enforces uniformity on
# the configuration.
##############################################################################
##############################################################################


##############################################################################
# Parameters
#
# Various parameters used in Py-ART.
##############################################################################

FILL_VALUE = fill_value  # the default fill value for masked arrays and
# the _FillValue key.

# The DEFAULT_FIELD_NAMES controls the field names which are used in
# by the correction and retrieval algorithms in Py-ART.  The keys of the
# dictionary are "internal" names which cannot change, the values are the field
# names which will be used in the algorithms by default.
# which will be used.  For best results use the names defined by the variables
# in simple configuration section which are also used in the DEFAULT_METADATA
# and FIELD_MAPPINGS variable.  If you choose to change a field name
# the names should also be changed in the DEFAULT_METADATA and FIELD_MAPPINGS
# variable. This is not required but highly suggested.

DEFAULT_FIELD_NAMES = {
    # internal field name (do not change): field name used (can change)
    "reflectivity": reflectivity,
    "corrected_reflectivity": corrected_reflectivity,
    "total_power": total_power,
    "velocity": velocity,
    "corrected_velocity": corrected_velocity,
    "spectrum_width": spectrum_width,
    "differential_reflectivity": differential_reflectivity,
    "corrected_differential_reflectivity": corrected_differential_reflectivity,
    "cross_correlation_ratio": cross_correlation_ratio,
    "normalized_coherent_power": normalized_coherent_power,
    "differential_phase": differential_phase,
    "unfolded_differential_phase": unfolded_differential_phase,
    "corrected_differential_phase": corrected_differential_phase,
    "specific_differential_phase": specific_differential_phase,
    "corrected_specific_differential_phase": corrected_specific_differential_phase,
    "linear_depolarization_ratio": linear_depolarization_ratio,
    "linear_depolarization_ratio_h": linear_depolarization_ratio_h,
    "linear_depolarization_ratio_v": linear_depolarization_ratio_v,
    "signal_to_noise_ratio": signal_to_noise_ratio,
    "rain_rate": rain_rate,
    "radar_estimated_rain_rate": radar_estimated_rain_rate,
    "radar_echo_classification": radar_echo_classification,
    "specific_attenuation": specific_attenuation,
}


##############################################################################
# Default metadata
#
# The DEFAULT_METADATA dictionary contains dictionaries which provide
# the default radar attribute and field metadata.  When reading in a file
# with Py-ART the FILE_SPECIFIC_METADATA variable is first queued for a
# metadata dictionary, if it is not found then the metadata in
# DEFAULT_METADATA is utilized.
##############################################################################

DEFAULT_METADATA = {
    # metadata for radar attributes
    # these closely follow the CF/Radial standard
    "azimuth": {
        "units": "foo",  # XXX changed
        "standard_name": "beam_azimuth_angle",
        "long_name": "azimuth_angle_from_true_north",
        "axis": "radial_azimuth_coordinate",
        "comment": "Azimuth of antenna relative to true north",
    },
    "elevation": {
        "units": "degrees",
        "standard_name": "beam_elevation_angle",
        "long_name": "elevation_angle_from_horizontal_plane",
        "axis": "radial_elevation_coordinate",
        "comment": "Elevation of antenna relative to the horizontal plane",
    },
    "range": {
        "units": "meters",
        "standard_name": "projection_range_coordinate",
        "long_name": "range_to_measurement_volume",
        "axis": "radial_range_coordinate",
        "spacing_is_constant": "true",
        "comment": ("Coordinate variable for range. Range to center of each bin."),
    },
    "time": {
        "units": "seconds",
        "standard_name": "time",
        "long_name": "time_in_seconds_since_volume_start",
        "calendar": "gregorian",
        "comment": (
            "Coordinate variable for time. "
            "Time at the center of each ray, in fractional seconds "
            "since the global variable time_coverage_start"
        ),
    },
    "sweep_mode": {
        "units": "unitless",
        "standard_name": "sweep_mode",
        "long_name": "Sweep mode",
        "comment": (
            'Options are: "sector", "coplane", "rhi", '
            '"vertical_pointing", "idle", "azimuth_surveillance", '
            '"elevation_surveillance", "sunscan", "pointing", '
            '"manual_ppi", "manual_rhi"'
        ),
    },
    "sweep_number": {
        "units": "count",
        "standard_name": "sweep_number",
        "long_name": "Sweep number",
    },
    "metadata": {
        "Conventions": "CF/Radial instrument_parameters",
        "version": "1.3",
        "title": "",
        "institution": "",
        "references": "",
        "source": "",
        "history": "",
        "comment": "",
        "instrument_name": "",
    },
    # metadata for radar sweep information dictionaries
    "sweep_start_ray_index": {
        "long_name": "Index of first ray in sweep, 0-based",
        "units": "count",
    },
    "sweep_end_ray_index": {
        "long_name": "Index of last ray in sweep, 0-based",
        "units": "count",
    },
    "fixed_angle": {
        "long_name": "Target angle for sweep",
        "units": "degrees",
        "standard_name": "target_fixed_angle",
    },
    # metadata for radar location attributes
    "latitude": {
        "long_name": "Latitude",
        "standard_name": "Latitude",
        "units": "degrees_north",
    },
    "longitude": {
        "long_name": "Longitude",
        "standard_name": "Longitude",
        "units": "degrees_east",
    },
    "altitude": {
        "long_name": "Altitude",
        "standard_name": "Altitude",
        "units": "meters",
        "positive": "up",
    },
    # metadata for instrument_parameter dictionary
    "prt_mode": {
        "comments": (
            'Pulsing mode Options are: "fixed", "staggered", '
            '"dual". Assumed "fixed" if missing.'
        ),
        "meta_group": "instrument_parameters",
        "long_name": "Pulsing mode",
        "units": "unitless",
    },
    "nyquist_velocity": {
        "units": "meters_per_second",
        "comments": "Unambiguous velocity",
        "meta_group": "instrument_parameters",
        "long_name": "Nyquist velocity",
    },
    "prt": {
        "units": "seconds",
        "comments": (
            "Pulse repetition time. For staggered prt, " "also see prt_ratio."
        ),
        "meta_group": "instrument_parameters",
        "long_name": "Pulse repetition time",
    },
    "unambiguous_range": {
        "units": "meters",
        "comments": "Unambiguous range",
        "meta_group": "instrument_parameters",
        "long_name": "Unambiguous range",
    },
    # reflectivity fields
    reflectivity: {
        "units": "dBZ",
        "standard_name": "equivalent_reflectivity_factor",
        "long_name": "Reflectivity",
        "coordinates": "elevation azimuth range",
    },
    corrected_reflectivity: {
        "units": "dBZ",
        "standard_name": "corrected_equivalent_reflectivity_factor",
        "long_name": "Corrected reflectivity",
        "coordinates": "elevation azimuth range",
    },
    total_power: {
        "units": "dBZ",
        "standard_name": "equivalent_reflectivity_factor",
        "long_name": "Total power",
        "coordinates": "elevation azimuth range",
    },
    # velocity fields
    velocity: {
        "units": "meters_per_second",
        "standard_name": "radial_velocity_of_scatterers_away_from_instrument",
        "long_name": "Mean dopper velocity",
        "coordinates": "elevation azimuth range",
    },
    corrected_velocity: {
        "units": "meters_per_second",
        "standard_name": (
            "corrected_radial_velocity_of_scatterers_away_from_instrument"
        ),
        "long_name": "Corrected mean doppler velocity",
        "coordinates": "elevation azimuth range",
    },
    # spectrum width fields
    spectrum_width: {
        "units": "meters_per_second",
        "standard_name": "doppler_spectrum_width",
        "long_name": "Doppler spectrum width",
        "coordinates": "elevation azimuth range",
    },
    # Dual Pol fields
    differential_reflectivity: {
        "units": "dB",
        "standard_name": "log_differential_reflectivity_hv",
        "long_name": "Differential reflectivity",
        "coordinates": "elevation azimuth range",
    },
    corrected_differential_reflectivity: {
        "units": "dB",
        "standard_name": "corrected_log_differential_reflectivity_hv",
        "long_name": "Corrected differential reflectivity",
        "coordinates": "elevation azimuth range",
    },
    cross_correlation_ratio: {
        "units": "ratio",
        "standard_name": "cross_correlation_ratio_hv",
        "long_name": "Cross correlation ratio (RHOHV)",
        "valid_max": 1.0,
        "valid_min": 0.0,
        "coordinates": "elevation azimuth range",
    },
    normalized_coherent_power: {
        "units": "ratio",
        "standard_name": "normalized_coherent_power",
        "long_name": "Normalized coherent power",
        "valid_max": 1.0,
        "valid_min": 0.0,
        "comment": "Also know as signal quality index (SQI)",
        "coordinates": "elevation azimuth range",
    },
    differential_phase: {
        "units": "degrees",
        "standard_name": "differential_phase_hv",
        "long_name": "Differential phase (PhiDP)",
        "valid_max": 180.0,
        "valid_min": -180.0,
        "coordinates": "elevation azimuth range",
    },
    unfolded_differential_phase: {
        "units": "degrees",
        "standard_name": "differential_phase_hv",
        "long_name": "Unfolded differential phase",
        "coordinates": "elevation azimuth range",
    },
    corrected_differential_phase: {
        "units": "degrees",
        "standard_name": "differential_phase_hv",
        "long_name": "Corrected differential phase",
        "coordinates": "elevation azimuth range",
    },
    specific_differential_phase: {
        "units": "degrees/km",
        "standard_name": "specific_differential_phase_hv",
        "long_name": "Specific differential phase (KDP)",
        "coordinates": "elevation azimuth range",
    },
    corrected_specific_differential_phase: {
        "units": "degrees/km",
        "standard_name": "specific_differential_phase_hv",
        "long_name": "Corrected specific differential phase (KDP)",
        "coordinates": "elevation azimuth range",
    },
    # depolarization ratio fields
    linear_depolarization_ratio: {
        "units": "dB",
        "standard_name": "log_linear_depolarization_ratio_hv",
        "long_name": "Linear depolarization ratio",
        "coordinates": "elevation azimuth range",
    },
    linear_depolarization_ratio_h: {
        "units": "dB",
        "standard_name": "log_linear_depolarization_ratio_h",
        "long_name": "Linear depolarization ratio horizontal",
        "coordinates": "elevation azimuth range",
    },
    linear_depolarization_ratio_v: {
        "units": "dB",
        "standard_name": "log_linear_depolarization_ratio_v",
        "long_name": "Linear depolarization ratio vertical",
        "coordinates": "elevation azimuth range",
    },
    # misc. fields
    signal_to_noise_ratio: {
        "units": "dB",
        "standard_name": "signal_to_noise_ratio",
        "long_name": "Signal to noise ratio",
        "coordinates": "elevation azimuth range",
    },
    rain_rate: {
        "units": "kg/m2/s",
        "standard_name": "rain_rate",
        "long_name": "Rain rate",
        "coordinates": "elevation azimuth range",
    },
    radar_estimated_rain_rate: {
        "units": "mm/hr",
        "standard_name": "radar_estimated_rain_rate",
        "long_name": "Radar estimated rain rate",
        "coordinates": "elevation azimuth range",
    },
    radar_echo_classification: {
        "units": "legend",
        "standard_name": "radar_echo_classification",
        "long_name": "Radar Echo classification",
        "coordinates": "elevation azimuth range",
    },
    specific_attenuation: {
        "units": "dB/km",
        "standard_name": "specific_attenuation",
        "long_name": "Specific attenuation",
        "valid_min": 0.0,
        "valid_max": 1.0,
        "coordinates": "elevation azimuth range",
    },
}


##############################################################################
# File specific metadata
#
# These dictionaries define metadata that is to be used only when reading in
# a given type of file.  This metadata is used in place of the
# DEFAULT_METADATA when it is avialable.  The main use of these variable
# is to define field specific data, it is safe to leave some/all of these
# empty if the default metadata is acceptable.
##############################################################################

# Metadata for Sigmet/IRIS files
sigmet_metadata = {}

# Metadata for NEXRAD Level II files (Archive and CDM files)
nexrad_metadata = {
    reflectivity: {
        "units": "dBZ",
        "standard_name": "equivalent_reflectivity_factor",
        "long_name": "Reflectivity",
        "valid_max": 94.5,
        "valid_min": -32.0,
        "coordinates": "elevation azimuth range",
    },
    velocity: {
        "units": "meters_per_second",
        "standard_name": "radial_velocity_of_scatterers_away_from_instrument",
        "long_name": "Mean doppler Velocity",
        "valid_max": 95.0,
        "valid_min": -95.0,
        "coordinates": "elevation azimuth range",
    },
    spectrum_width: {
        "units": "meters_per_second",
        "standard_name": "doppler_spectrum_width",
        "long_name": "Spectrum Width",
        "valid_max": 63.0,
        "valid_min": -63.5,
        "coordinates": "elevation azimuth range",
    },
    differential_reflectivity: {
        "units": "dB",
        "standard_name": "log_differential_reflectivity_hv",
        "long_name": "log_differential_reflectivity_hv",
        "valid_max": 7.9375,
        "valid_min": -7.8750,
        "coordinates": "elevation azimuth range",
    },
    differential_phase: {
        "units": "degrees",
        "standard_name": "differential_phase_hv",
        "long_name": "differential_phase_hv",
        "valid_max": 360.0,
        "valid_min": 0.0,
        "coordinates": "elevation azimuth range",
    },
    cross_correlation_ratio: {
        "units": "ratio",
        "standard_name": "cross_correlation_ratio_hv",
        "long_name": "Cross correlation_ratio (RHOHV)",
        "valid_max": 1.0,
        "valid_min": 0.0,
        "coordinates": "elevation azimuth range",
    },
}

# Metadata for CF/Radial files
cfradial_metadata = {}

# Metadata for MDV files
mdv_metadata = {"time": {"foo": "bar"}}  # XXX changed

# Metadata for RSL files
rsl_metadata = {}

FILE_SPECIFIC_METADATA = {  # Required
    "sigmet": sigmet_metadata,
    "nexrad_archive": nexrad_metadata,
    "nexrad_cdm": nexrad_metadata,
    "cfradial": cfradial_metadata,
    "mdv": mdv_metadata,
    "rsl": rsl_metadata,
}

##############################################################################
# Field name mapping
#
# These dictionaries map file field names or data types to a radar field
# name.  These are used to populate the radar.fields dictionary during a read
# in Py-ART.  A value of None will not include that field in the radar object.
# These can be over-ridden on a per-read basis using the field_mapping
# parameter, or using setting the file_field_names parameter to True.
##############################################################################

# Sigmet/IRIS file field mapping
# Note that multiple sigmet fields map to the same radar field, if
# more than one of these fields are present the radar field will be
# overwritten with the last sigmet field.
sigmet_field_mapping = {
    # Sigmet data type :field name              # (Data_type) Description
    "XHDR": None,  # (0) Extended Header
    "DBT": total_power,  # (1) Total Power
    "DBZ": reflectivity,  # (2) Reflectivity
    "VEL": velocity,  # (3) Velocity
    "WIDTH": spectrum_width,  # (4) Width
    "ZDR": differential_reflectivity,  # (5) Diff. reflectivity
    "DBZC": corrected_reflectivity,  # (7) Corrected reflectivity
    "DBT2": total_power,  # (8) Total Power
    "DBZ2": reflectivity,  # (9) Reflectivity
    "VEL2": velocity,  # (10) Velocity
    "WIDTH2": spectrum_width,  # (11) Width
    "ZDR2": differential_reflectivity,  # (12) Diff. reflectivity
    "RAINRATE2": radar_estimated_rain_rate,  # (13) Rainfall rate
    "KDP": specific_differential_phase,  # (14) KDP (diff. phase)
    "KDP2": specific_differential_phase,  # (15) KDP (diff. phase)
    "PHIDP": differential_phase,  # (16) PhiDP (diff. phase)
    "VELC": corrected_velocity,  # (17) Corrected velocity
    "SQI": normalized_coherent_power,  # (18) SQI
    "RHOHV": cross_correlation_ratio,  # (19) RhoHV
    "RHOHV2": cross_correlation_ratio,  # (20) RhoHV
    "DBZC2": corrected_reflectivity,  # (21) Corrected Reflec.
    "VELC2": corrected_velocity,  # (21) Corrected Velocity
    "SQI2": normalized_coherent_power,  # (23) SQI
    "PHIDP2": differential_phase,  # (24) PhiDP (diff. phase)
    "LDRH": linear_depolarization_ratio_h,  # (25) LDR xmt H, rcv V
    "LDRH2": linear_depolarization_ratio_h,  # (26) LDR xmt H, rcv V
    "LDRV": linear_depolarization_ratio_v,  # (27) LDR xmt V, rcv H
    "LDRV2": linear_depolarization_ratio_v,  # (28) LDR xmt V, rcv H
    "HEIGHT": None,  # (32) Height (1/10 km)
    "VIL2": None,  # (33) Linear Liquid
    "RAW": None,  # (34) Raw Data
    "SHEAR": None,  # (35) Wind Shear
    "DIVERGE2": None,  # (36) Divergence
    "FLIQUID2": None,  # (37) Floated liquid
    "USER": None,  # (38) User type
    "OTHER": None,  # (39) Unspecified
    "DEFORM2": None,  # (40) Deformation
    "VVEL2": None,  # (41) Vertical velocity
    "HVEL2": None,  # (42) Horizontal velocity
    "HDIR2": None,  # (43) Horiz. wind direction
    "AXDIL2": None,  # (44) Axis of dilation
    "TIME2": None,  # (45) Time in seconds
    "RHOH": None,  # (46) Rho, xmt H, rcv V
    "RHOH2": None,  # (47) Rho, xmt H, rcv V
    "RHOV": None,  # (48) Rho, xmt V, rcv H
    "RHOV2": None,  # (49) Rho, xmt V, rcv H
    "PHIH": None,  # (50) Phi, xmt H, rcv V
    "PHIH2": None,  # (51) Phi, xmt H, rcv V
    "PHIV": None,  # (52) Phi, xmt V, rcv H
    "PHIV2": None,  # (53) Phi, xmt V, rcv H
    "USER2": None,  # (54) User type
    "HCLASS": radar_echo_classification,  # (55) Hydrometeor class
    "HCLASS2": radar_echo_classification,  # (56) Hydrometeor class
    "ZDRC": corrected_differential_reflectivity,
    # (57) Corrected diff. refl.
    "ZDRC2": corrected_differential_reflectivity,
    # (58) Corrected diff. refl.
}


# NEXRAD Level II Archive files
nexrad_archive_field_mapping = {
    # NEXRAD field: radar field name
    "REF": reflectivity,
    "VEL": velocity,
    "SW": spectrum_width,
    "ZDR": differential_reflectivity,
    "PHI": differential_phase,
    "RHO": cross_correlation_ratio,
}

# NEXRAD Level II CDM files
nexrad_cdm_field_mapping = {
    # CDM variable name (without _HI): radar field name
    "Reflectivity": reflectivity,
    "RadialVelocity": velocity,
    "SpectrumWidth": spectrum_width,
    "DifferentialReflectivity": differential_reflectivity,
    "DifferentialPhase": differential_phase,
    "CorrelationCoefficient": cross_correlation_ratio,
}

# MDV files
mdv_field_mapping = {
    # MDV moment: radar field name
    "DBZ_F": velocity,  # XXX changed
    "VEL_F": velocity,
    "WIDTH_F": spectrum_width,
    "ZDR_F": differential_reflectivity,
    "RHOHV_F": cross_correlation_ratio,
    "NCP_F": normalized_coherent_power,
    "KDP_F": specific_differential_phase,
    "PHIDP_F": differential_phase,
    "VEL_COR": corrected_velocity,
    "PHIDP_UNF": unfolded_differential_phase,
    "KDP_SOB": corrected_specific_differential_phase,
    "DBZ_AC": corrected_reflectivity,
}

# CF/Radial files
cfradial_field_mapping = {}

# RSL files
# Note that multiple RSL field map to the same radar field, if
# more than one of these fields are present in the RSL data structure
# the radar field will be overwritten with the last field.
rsl_field_mapping = {
    # RSL 2 letter field: radar field           # RSL description
    "DZ": reflectivity,  # reflectivity
    "VR": velocity,  # velocity
    "SW": spectrum_width,  # spectrum width
    "CZ": corrected_reflectivity,  # corrected reflectivity
    "ZT": reflectivity,  # uncorrected reflectivity
    "DR": differential_reflectivity,  # differential reflectivity
    "LR": differential_reflectivity,  # another diff. reflectivity
    "ZD": differential_reflectivity,  # another diff. reflectivity
    "DM": None,  # received power
    "RH": cross_correlation_ratio,  # RhoHV
    "PH": differential_phase,  # PhiDP
    "XZ": None,  # X-band reflectivity
    "CD": corrected_differential_reflectivity,  # Corrected DR.
    "MZ": None,  # DZ mask
    "MD": None,  # DR Mask
    "ZE": corrected_reflectivity,  # edited reflectivity
    "VE": corrected_velocity,  # edited velocity
    "KD": specific_differential_phase,  # specific diff. phase
    "TI": None,  # TIME (unknown)
    "DX": None,  # ???
    "CH": None,  # ???
    "AH": None,  # ???
    "CV": None,  # ???
    "AV": None,  # ???
    "SQ": normalized_coherent_power,  # Signal Quality Index (sigmet)
    "VS": None,  # Radial Vel. combined
    "VL": None,  # Radial Vel. combined
    "VG": None,  # Radial Vel. combined
    "VT": None,  # Radial Vel. combined
    "NP": normalized_coherent_power,  # Normalized Coherent Power
    "HC": radar_echo_classification,  # Hydroclass
    "VC": None,  # Radial Vel. Corrected.
    "V2": None,  # Radial Vel cut 2
    "S2": None,  # Spectrum width cut 2
    "V3": None,  # Radial Vel cut 3
    "S3": None,  # Spectrum width cut 3
}

FIELD_MAPPINGS = {  # Required variable
    "sigmet": sigmet_field_mapping,
    "nexrad_archive": nexrad_archive_field_mapping,
    "nexrad_cdm": nexrad_cdm_field_mapping,
    "cfradial": cfradial_field_mapping,
    "mdv": mdv_field_mapping,
    "rsl": rsl_field_mapping,
}



from copy import deepcopy

import numpy as np

import pyart


def test_gatemapper():
    # Make fake radar with target
    old_radar = pyart.testing.make_target_radar()
    new_radar = deepcopy(old_radar)
    new_radar.latitude["data"] = old_radar.latitude["data"] + 0.001
    new_radar.longitude["data"] = old_radar.longitude["data"] + 0.001
    gate_mapper = pyart.map.GateMapper(new_radar, old_radar)
    mapped_radar = gate_mapper.mapped_radar(["reflectivity"])

    # Test point outside of 1 min tolerance
    assert gate_mapper[20, 20] == (None, None)
    assert gate_mapper[4, 4] == (26, 11)
    assert (
        mapped_radar.fields["reflectivity"]["data"][26, 11]
        == old_radar.fields["reflectivity"]["data"][4, 4]
    )


def test_gatemapper_gatefilter():
    # Make fake radar with target
    old_radar = pyart.testing.make_target_radar()
    new_radar = deepcopy(old_radar)
    new_radar.latitude["data"] = old_radar.latitude["data"] + 0.001
    new_radar.longitude["data"] = old_radar.longitude["data"] + 0.001
    gatefilter = pyart.filters.GateFilter(old_radar)
    gatefilter.exclude_below("reflectivity", 40)
    gate_mapper = pyart.map.GateMapper(new_radar, old_radar, gatefilter_src=gatefilter)
    mapped_radar = gate_mapper.mapped_radar(["reflectivity"])
    assert gate_mapper[4, 4] == (26, 11)
    assert np.ma.is_masked(mapped_radar.fields["reflectivity"]["data"][26, 11])



"""
Sample radar files in a number of formats. Many of these files
are incomplete, they should only be used for testing, not production.

MDV_PPI_FILE
MDV_RHI_FILE
CFRADIAL_PPI_FILE
CFRADIAL_RHI_FILE
CFRADIAL_CR_RASTER_FILE
CHL_RHI_FILE
SIGMET_PPI_FILE
SIGMET_RHI_FILE
NEXRAD_ARCHIVE_MSG31_FILE
NEXRAD_ARCHIVE_MSG31_COMPRESSED_FILE
NEXRAD_ARCHIVE_MSG1_FILE
NEXRAD_LEVEL3_MSG19
NEXRAD_LEVEL3_MSG163
NEXRAD_CDM_FILE
UF_FILE
INTERP_SOUNDE_FILE

"""

import os

DATA_PATH = os.path.join(os.path.dirname(__file__), "data")

MDV_PPI_FILE = os.path.join(DATA_PATH, "example_mdv_ppi.mdv")
MDV_RHI_FILE = os.path.join(DATA_PATH, "example_mdv_rhi.mdv")
MDV_GRID_FILE = os.path.join(DATA_PATH, "example_mdv_grid.mdv")
CFRADIAL_PPI_FILE = os.path.join(DATA_PATH, "example_cfradial_ppi.nc")
CFRADIAL_RHI_FILE = os.path.join(DATA_PATH, "example_cfradial_rhi.nc")
CFRADIAL_CR_RASTER_FILE = os.path.join(DATA_PATH, "example_cfradial_cr_raster.nc")
CHL_RHI_FILE = os.path.join(DATA_PATH, "example_chl_rhi.chl")
SIGMET_PPI_FILE = os.path.join(DATA_PATH, "example_sigmet_ppi.sigmet")
SIGMET_RHI_FILE = os.path.join(DATA_PATH, "example_sigmet_rhi.sigmet")
NEXRAD_ARCHIVE_MSG31_FILE = os.path.join(DATA_PATH, "example_nexrad_archive_msg31.bz2")
NEXRAD_ARCHIVE_MSG31_COMPRESSED_FILE = os.path.join(
    DATA_PATH, "example_nexrad_archive_msg31_compressed.ar2v"
)
NEXRAD_ARCHIVE_MSG1_FILE = os.path.join(DATA_PATH, "example_nexrad_archive_msg1.bz2")
# NEXRAD Level 3 file downloaded from NCDC with filenames:
# KBMX_SDUS54_N0RBMX_201501020205
# KBMX_SDUS84_N0KBMX_201501020205
NEXRAD_LEVEL3_MSG19 = os.path.join(DATA_PATH, "example_nexrad_level3_msg19")
NEXRAD_LEVEL3_MSG163 = os.path.join(DATA_PATH, "example_nexrad_level3_msg163")
NEXRAD_LEVEL3_MSG176 = os.path.join(DATA_PATH, "example_nexrad_level3_msg176")
NEXRAD_CDM_FILE = os.path.join(DATA_PATH, "example_nexrad_cdm.bz2")
UF_FILE = os.path.join(DATA_PATH, "example_uf_ppi.uf")
INTERP_SOUNDE_FILE = os.path.join(DATA_PATH, "example_interpolatedsonde.cdf")
SONDE_FILE = os.path.join(DATA_PATH, "example_arm_sonde.cdf")
_EXAMPLE_RAYS_FILE = os.path.join(DATA_PATH, "example_rays.npz")



#! /usr/bin/env python
"""
Extract a single ray from a C-SAPR PPI volume and store in a .npz file.

"""

import numpy as np

import pyart

# read in the full radar
radar = pyart.io.read("095636.mdv")

# extract ray 191 from the requested fields
fields_to_extract = [
    "reflectivity",
    "normalized_coherent_power",
    "cross_correlation_ratio",
    "specific_differential_phase",
    "differential_phase",
    "differential_reflectivity",
]
ray_dic = dict()
for field_name in fields_to_extract:
    ray = radar.fields[field_name]["data"].data[191]
    ray = ray.reshape(1, -1)
    ray_dic[field_name] = ray

# save into a npz file
np.savez("example_rays.npz", **ray_dic)



"""
Configuration file for the Python ARM Radar Toolkit (Py-ART)

The values for a number of Py-ART parameters and the default metadata created
when reading files, correcting fields, etc. is controlled by this single
Python configuration file.

Py-ART's configuration can be modified by setting the environment variable
PYART_CONFIG to point to a configuration file with formatting similar to this
file.

The recommended method for changing these defaults is for users to copy this
file into their home directory, rename it to .pyart_config.py, make any
desired changes, and adjust their login scripts to set the PYART_CONFIG
environment variable to point to .pyart_config.py in their home directory.

Py-ART's configuration can also be modified within a script or shell session
using the load_config functions, such modification will last until the end of
the script/session or until a new configuration is loaded.

"""

##############################################################################
##############################################################################
# Simple configuration
#
# Adjust the values of the variable (right hand side of the equal sign) in
# this section for an easy method of customizing Py-ART.  Do not change the
# variable names (the left hand side of the equal sign). More advanced
# settings are based upon these variables.  Most users will find that
# adjusting this section is all that is needed.
##############################################################################
##############################################################################

# The default fill value for masked arrays and _FillValue keys
fill_value = -9999.0

# Field names used when reading in radar files and in the various correction
# and retrieval algorithms. The comments in this section provide additional
# information about the fields in that section.

# Radar reflectivity fields, DZ
reflectivity = "reflectivity"
corrected_reflectivity = "corrected_reflectivity"
reflectivity_texture = "reflectivity_texture"
total_power = "total_power"

# Mean Doppler velocity fields, VEL
velocity = "velocity"
corrected_velocity = "corrected_velocity"
simulated_velocity = "simulated_velocity"

# Spectral width fields, SW
spectrum_width = "spectrum_width"

# Differential reflectivity fields, ZDR
differential_reflectivity = "differential_reflectivity"
corrected_differential_reflectivity = "corrected_differential_reflectivity"
differential_reflectivity_texture = "differential_reflectivity_texture"

# Cross correlation ratio, correlation coefficient, RhoHV
cross_correlation_ratio = "cross_correlation_ratio"
cross_correlation_ratio_texture = "cross_correlation_ratio_texture"

# Normalized coherent power, signal quality index, SQI, NCP
normalized_coherent_power = "normalized_coherent_power"
logarithmic_cross_correlation_ratio = "logarithmic_cross_correlation_ratio"

# Differential phase shift, PhiDP
differential_phase = "differential_phase"
unfolded_differential_phase = "unfolded_differential_phase"
corrected_differential_phase = "corrected_differential_phase"
differential_phase_texture = "differential_phase_texture"

# Specific differential phase shift, KDP
specific_differential_phase = "specific_differential_phase"
corrected_specific_differential_phase = "corrected_specific_differential_phase"

# Linear depolarization ratio (h - horizontal, v - vertical), LDR
linear_depolarization_ratio = "linear_polarization_ratio"
linear_depolarization_ratio_h = "linear_depolarization_ratio_h"
linear_depolarization_ratio_v = "linear_depolarization_ratio_v"

# Circular depolarization ratio
circular_depolarization_ratio = "circular_depolarization_ratio"

# Misc fields
signal_to_noise_ratio = "signal_to_noise_ratio"
noisedBZ_hh = "noisedBZ_hh"
noisedBZ_vv = "noisedBZ_vv"
rain_rate = "rain_rate"
radar_estimated_rain_rate = "radar_estimated_rain_rate"
radar_echo_classification = "radar_echo_classification"
specific_attenuation = "specific_attenuation"
specific_differential_attenuation = "specific_differential_attenuation"
clutter_filter_power_removed = "clutter_filter_power_removed"

# Textures
differential_phase_texture = "differential_phase_texture"

# Wind retrieval fields
eastward_wind_component = "eastward_wind_component"
northward_wind_component = "northward_wind_component"
vertical_wind_component = "vertical_wind_component"

# profile variables
height = "height"
interpolated_profile = "interpolated_profile"
height_over_iso0 = "height_over_iso0"
temperature = "temperature"

# path integrated attenuation
path_integrated_attenuation = "path_integrated_attenuation"
path_integrated_differential_attenuation = (
    "path_integrated" + "differential_attenuation"
)

# End of Simple Configuration section

##############################################################################
##############################################################################
# Advanced Configuration
#
# Most users will not want to make any changes in this section.  For users
# who want a more fine-grained control over Py-ART's configuration this
# section provides access to these controls.  The layout of this section can
# be changed, the only requirement for a valid configuration file is that
# the ALL CAPITALIZED variable must must be present with the formatting
# present in this file.  These required variables are:
#
# FILL_VALUE, DEFAULT_METADATA, FILE_SPECIFIC_METADATA, FIELD_MAPPINGS,
# DEFAULT_FIELD_NAMES
#
# This section makes generous use of the variables in the Simple Configuration
# section, this is not required, but simplifies and enforces uniformity on
# the configuration.
##############################################################################
##############################################################################


##############################################################################
# Parameters
#
# Various parameters used in Py-ART.
##############################################################################

# the default fill value for masked arrays and the _FillValue key
FILL_VALUE = fill_value

# The DEFAULT_FIELD_NAMES controls the field names which are used in the
# correction and retrieval algorithms in Py-ART. The keys of the dictionary
# are "internal" names which cannot change, the values are the field names
# which will be used in the algorithms by default. For best results use the
# names defined by the variables in simple configuration section which are
# also used in the DEFAULT_METADATA and FIELD_MAPPINGS variable. If you
# choose to change a field name the names should also be changed in the
# DEFAULT_METADATA and FIELD_MAPPINGS variable. This is not required but
# highly suggested.

DEFAULT_FIELD_NAMES = {
    # Internal field name (do not change): field name used (can change)
    "reflectivity": reflectivity,
    "corrected_reflectivity": corrected_reflectivity,
    "total_power": total_power,
    "velocity": velocity,
    "corrected_velocity": corrected_velocity,
    "simulated_velocity": simulated_velocity,
    "spectrum_width": spectrum_width,
    "differential_reflectivity": differential_reflectivity,
    "corrected_differential_reflectivity": corrected_differential_reflectivity,
    "cross_correlation_ratio": cross_correlation_ratio,
    "logarithmic_cross_correlation_ratio": logarithmic_cross_correlation_ratio,
    "normalized_coherent_power": normalized_coherent_power,
    "differential_phase": differential_phase,
    "unfolded_differential_phase": unfolded_differential_phase,
    "corrected_differential_phase": corrected_differential_phase,
    "specific_differential_phase": specific_differential_phase,
    "corrected_specific_differential_phase": corrected_specific_differential_phase,
    "linear_depolarization_ratio": linear_depolarization_ratio,
    "linear_depolarization_ratio_h": linear_depolarization_ratio_h,
    "linear_depolarization_ratio_v": linear_depolarization_ratio_v,
    "circular_depolarization_ratio": circular_depolarization_ratio,
    "signal_to_noise_ratio": signal_to_noise_ratio,
    "noisedBZ_hh": noisedBZ_hh,
    "noisedBZ_vv": noisedBZ_vv,
    "rain_rate": rain_rate,
    "radar_estimated_rain_rate": radar_estimated_rain_rate,
    "radar_echo_classification": radar_echo_classification,
    "specific_attenuation": specific_attenuation,
    "differential_phase_texture": differential_phase_texture,
    "eastward_wind_component": eastward_wind_component,
    "northward_wind_component": northward_wind_component,
    "vertical_wind_component": vertical_wind_component,
    "height": height,
    "height_over_iso0": height_over_iso0,
    "interpolated_profile": interpolated_profile,
    "temperature": temperature,
    "path_integrated_attenuation": path_integrated_attenuation,
    "specific_differential_attenuation": specific_differential_attenuation,
    "path_integrated_differential_attenuation": path_integrated_differential_attenuation,
    "clutter_filter_power_removed": clutter_filter_power_removed,
    "reflectivity_texture": reflectivity_texture,
    "differential_reflectivity_texture": differential_reflectivity_texture,
    "cross_correlation_ratio_texture": cross_correlation_ratio_texture,
    "differential_phase_texture": differential_phase_texture,
}


##############################################################################
# Default metadata
#
# The DEFAULT_METADATA dictionary contains dictionaries which provide the
# default radar attribute and field metadata. When reading in a file with
# Py-ART the FILE_SPECIFIC_METADATA variable is first queued for a metadata
# dictionary, if it is not found then the metadata in DEFAULT_METADATA is
# utilized.
##############################################################################

DEFAULT_METADATA = {
    # Metadata for radar attributes. These closely follow the CF/Radial
    # standard
    "azimuth": {
        "units": "degrees",
        "standard_name": "beam_azimuth_angle",
        "long_name": "azimuth_angle_from_true_north",
        "axis": "radial_azimuth_coordinate",
        "comment": "Azimuth of antenna relative to true north",
    },
    "elevation": {
        "units": "degrees",
        "standard_name": "beam_elevation_angle",
        "long_name": "elevation_angle_from_horizontal_plane",
        "axis": "radial_elevation_coordinate",
        "comment": "Elevation of antenna relative to the horizontal plane",
    },
    "scan_rate": {
        "units": "degrees_per_second",
        "long_name": "Antenna angle scan rate",
    },
    "range": {
        "units": "meters",
        "standard_name": "projection_range_coordinate",
        "long_name": "range_to_measurement_volume",
        "axis": "radial_range_coordinate",
        "spacing_is_constant": "true",
        "comment": ("Coordinate variable for range. Range to center of each bin."),
    },
    "time": {
        "units": "seconds",
        "standard_name": "time",
        "long_name": "time_in_seconds_since_volume_start",
        "calendar": "gregorian",
        "comment": (
            "Coordinate variable for time. "
            "Time at the center of each ray, in fractional seconds "
            "since the global variable time_coverage_start"
        ),
    },
    "metadata": {
        "Conventions": "CF/Radial instrument_parameters",
        "version": "1.3",
        "title": "",
        "institution": "",
        "references": "",
        "source": "",
        "history": "",
        "comment": "",
        "instrument_name": "",
    },
    # Metadata for radar sweep information dictionaries
    "sweep_number": {
        "units": "count",
        "standard_name": "sweep_number",
        "long_name": "Sweep number",
    },
    "sweep_mode": {
        "units": "unitless",
        "standard_name": "sweep_mode",
        "long_name": "Sweep mode",
        "comment": (
            'Options are: "sector", "coplane", "rhi", '
            '"vertical_pointing", "idle", "azimuth_surveillance", '
            '"elevation_surveillance", "sunscan", "pointing", '
            '"manual_ppi", "manual_rhi"'
        ),
    },
    "fixed_angle": {
        "long_name": "Target angle for sweep",
        "units": "degrees",
        "standard_name": "target_fixed_angle",
    },
    "sweep_start_ray_index": {
        "long_name": "Index of first ray in sweep, 0-based",
        "units": "count",
    },
    "sweep_end_ray_index": {
        "long_name": "Index of last ray in sweep, 0-based",
        "units": "count",
    },
    "rays_per_sweep": {"long_name": "Number of rays in each sweep", "units": "count"},
    "target_scan_rate": {
        "long_name": "Target scan rate for sweep",
        "units": "degrees_per_second",
    },
    "rays_are_indexed": {
        "long_name": "Flag for indexed rays",
        "units": "unitless",
        "options": (
            "true: rays are indexed to a regular grid, "
            + "false: rays are not indexed to a regular grid"
        ),
    },
    "ray_angle_res": {
        "long_name": "Angular resolution between rays",
        "units": "degrees",
        "comment": "Only applicable when rays_are_indexed variable is true",
    },
    # Metadata for radar location attributes
    "latitude": {
        "long_name": "Latitude",
        "standard_name": "Latitude",
        "units": "degrees_north",
    },
    "longitude": {
        "long_name": "Longitude",
        "standard_name": "Longitude",
        "units": "degrees_east",
    },
    "altitude": {
        "long_name": "Altitude",
        "standard_name": "Altitude",
        "units": "meters",
        "positive": "up",
    },
    "gate_x": {
        "long_name": "Cartesian x location of gate with origin at the radar",
        "units": "meters",
    },
    "gate_y": {
        "long_name": "Cartesian y location of gate with origin at the radar",
        "units": "meters",
    },
    "gate_z": {
        "long_name": "Cartesian z location of gate with origin at the radar",
        "units": "meters",
    },
    "gate_edge_x": {
        "long_name": "Cartesian x location of the edges of each gate",
        "units": "meters",
    },
    "gate_edge_y": {
        "long_name": "Cartesian y location of the edges of each gate",
        "units": "meters",
    },
    "gate_edge_z": {
        "long_name": "Cartesian z location of the edges of each gate",
        "units": "meters",
    },
    "gate_longitude": {
        "long_name": "Longitude of radar gate.",
        "units": "degrees_north",
    },
    "gate_latitude": {"long_name": "Latitude of radar gate", "units": "degrees_east"},
    "gate_altitude": {"long_name": "Altitude of radar gate", "units": "meters"},
    # Metadata for instrument_parameter dictionary
    "prt_mode": {
        "comments": (
            'Pulsing mode Options are: "fixed", "staggered", '
            '"dual". Assumed "fixed" if missing.'
        ),
        "meta_group": "instrument_parameters",
        "long_name": "Pulsing mode",
        "units": "unitless",
    },
    "nyquist_velocity": {
        "units": "meters_per_second",
        "comments": "Unambiguous velocity",
        "meta_group": "instrument_parameters",
        "long_name": "Nyquist velocity",
    },
    "prt": {
        "units": "seconds",
        "comments": (
            "Pulse repetition time. For staggered prt, " "also see prt_ratio."
        ),
        "meta_group": "instrument_parameters",
        "long_name": "Pulse repetition time",
    },
    "unambiguous_range": {
        "units": "meters",
        "comments": "Unambiguous range",
        "meta_group": "instrument_parameters",
        "long_name": "Unambiguous range",
    },
    "pulse_width": {
        "units": "seconds",
        "comments": "Pulse width",
        "meta_group": "instrument_parameters",
        "long_name": "Pulse width",
    },
    "prt_ratio": {
        "units": "unitless",
        "meta_group": "instrument_parameters",
        "long_name": "Pulse repetition frequency ratio",
    },
    "frequency": {
        "units": "s-1",
        "meta_group": "instrument_parameters",
        "long_name": "Radiation frequency",
    },
    "n_samples": {
        "units": "unitless",
        "meta_group": "instrument_parameters",
        "long_name": "Number of samples used to compute moments",
    },
    # non-standard parameter for specifying the PRF high/low for each ray
    "prf_flag": {
        "units": "unitless",
        "comments": "PRF used to collect ray. 0 for high PRF, 1 for low PRF.",
        "meta_group": "instrument_parameters",
        "long_name": "PRF flag",
    },
    # Metadata for radar_parameter sub-convention
    "radar_beam_width_h": {
        "units": "degrees",
        "meta_group": "radar_parameters",
        "long_name": "Antenna beam width H polarization",
    },
    "radar_beam_width_v": {
        "units": "degrees",
        "meta_group": "radar_parameters",
        "long_name": "Antenna beam width V polarization",
    },
    # Metadata for airborne radar parameters
    "rotation": {
        "units": "degrees",
        "standard_name": "ray_rotation_angle_relative_to_platform",
        "long_name": "Ray rotation angle relative to platform",
    },
    "tilt": {
        "units": "degrees",
        "standard_name": "ray_tilt_angle_relative_to_platform",
        "long_name": "Ray tilt angle relative to platform",
    },
    "roll": {
        "units": "degrees",
        "standard_name": "platform_roll_angle",
        "long_name": "Platform roll angle",
    },
    "drift": {
        "units": "degrees",
        "standard_name": "platform_drift_angle",
        "long_name": "Platform drift angle",
    },
    "heading": {
        "units": "degrees",
        "standard_name": "platform_heading_angle",
        "long_name": "Platform heading angle",
    },
    "pitch": {
        "units": "degrees",
        "standard_name": "platform_pitch_angle",
        "long_name": "Platform pitch angle",
    },
    "georefs_applied": {
        "units": "unitless",
        "standard_name": "georefs_have_been_applied_to_ray",
        "long_name": "Geoferences have been applied to ray",
        "comment": "1 if georefs have been applied, 0 otherwise",
    },
    # Reflectivity fields
    reflectivity: {
        "units": "dBZ",
        "standard_name": "equivalent_reflectivity_factor",
        "long_name": "Reflectivity",
        "coordinates": "elevation azimuth range",
    },
    corrected_reflectivity: {
        "units": "dBZ",
        "standard_name": "corrected_equivalent_reflectivity_factor",
        "long_name": "Corrected reflectivity",
        "coordinates": "elevation azimuth range",
    },
    total_power: {
        "units": "dBZ",
        "standard_name": "equivalent_reflectivity_factor",
        "long_name": "Total power",
        "coordinates": "elevation azimuth range",
    },
    # Velocity fields
    velocity: {
        "units": "meters_per_second",
        "standard_name": "radial_velocity_of_scatterers_away_from_instrument",
        "long_name": "Mean dopper velocity",
        "coordinates": "elevation azimuth range",
    },
    corrected_velocity: {
        "units": "meters_per_second",
        "standard_name": (
            "corrected_radial_velocity_of_scatterers_away_from_instrument"
        ),
        "long_name": "Corrected mean doppler velocity",
        "coordinates": "elevation azimuth range",
    },
    simulated_velocity: {
        "units": "meters_per_second",
        "standard_name": ("radial_velocity_of_scatterers_away_from_instrument"),
        "long_name": "Simulated mean doppler velocity",
        "coordinates": "elevation azimuth range",
    },
    # Spectrum width fields
    spectrum_width: {
        "units": "meters_per_second",
        "standard_name": "doppler_spectrum_width",
        "long_name": "Doppler spectrum width",
        "coordinates": "elevation azimuth range",
    },
    # Dual-polarization fields
    differential_reflectivity: {
        "units": "dB",
        "standard_name": "log_differential_reflectivity_hv",
        "long_name": "Differential reflectivity",
        "coordinates": "elevation azimuth range",
    },
    corrected_differential_reflectivity: {
        "units": "dB",
        "standard_name": "corrected_log_differential_reflectivity_hv",
        "long_name": "Corrected differential reflectivity",
        "coordinates": "elevation azimuth range",
    },
    cross_correlation_ratio: {
        "units": "ratio",
        "standard_name": "cross_correlation_ratio_hv",
        "long_name": "Cross correlation ratio (RHOHV)",
        "valid_max": 1.0,
        "valid_min": 0.0,
        "coordinates": "elevation azimuth range",
    },
    logarithmic_cross_correlation_ratio: {
        "units": "ratio",
        "long_name": "Logarithmic cross correlation ratio",
        "coordinates": "elevation azimuth range",
    },
    normalized_coherent_power: {
        "units": "ratio",
        "standard_name": "normalized_coherent_power",
        "long_name": "Normalized coherent power",
        "valid_max": 1.0,
        "valid_min": 0.0,
        "comment": "Also know as signal quality index (SQI)",
        "coordinates": "elevation azimuth range",
    },
    differential_phase: {
        "units": "degrees",
        "standard_name": "differential_phase_hv",
        "long_name": "Differential phase (PhiDP)",
        "valid_max": 180.0,
        "valid_min": -180.0,
        "coordinates": "elevation azimuth range",
    },
    unfolded_differential_phase: {
        "units": "degrees",
        "standard_name": "differential_phase_hv",
        "long_name": "Unfolded differential phase",
        "coordinates": "elevation azimuth range",
    },
    corrected_differential_phase: {
        "units": "degrees",
        "standard_name": "differential_phase_hv",
        "long_name": "Corrected differential phase",
        "coordinates": "elevation azimuth range",
    },
    specific_differential_phase: {
        "units": "degrees/km",
        "standard_name": "specific_differential_phase_hv",
        "long_name": "Specific differential phase (KDP)",
        "coordinates": "elevation azimuth range",
    },
    corrected_specific_differential_phase: {
        "units": "degrees/km",
        "standard_name": "specific_differential_phase_hv",
        "long_name": "Corrected specific differential phase (KDP)",
        "coordinates": "elevation azimuth range",
    },
    # Depolarization ratio fields
    linear_depolarization_ratio: {
        "units": "dB",
        "standard_name": "log_linear_depolarization_ratio_hv",
        "long_name": "Linear depolarization ratio",
        "coordinates": "elevation azimuth range",
    },
    linear_depolarization_ratio_h: {
        "units": "dB",
        "standard_name": "log_linear_depolarization_ratio_h",
        "long_name": "Linear depolarization ratio horizontal",
        "coordinates": "elevation azimuth range",
    },
    linear_depolarization_ratio_v: {
        "units": "dB",
        "standard_name": "log_linear_depolarization_ratio_v",
        "long_name": "Linear depolarization ratio vertical",
        "coordinates": "elevation azimuth range",
    },
    circular_depolarization_ratio: {
        "units": "dB",
        "long_name": "Circular depolarization ratio",
        "coordinates": "elevation azimuth range",
    },
    # Misc fields
    signal_to_noise_ratio: {
        "units": "dB",
        "standard_name": "signal_to_noise_ratio",
        "long_name": "Signal to noise ratio",
        "coordinates": "elevation azimuth range",
    },
    noisedBZ_hh: {
        "units": "dBZ",
        "long_name": "noise in dBZ horizontal",
        "coordinates": "elevation azimuth range",
    },
    noisedBZ_vv: {
        "units": "dBZ",
        "long_name": "noise in dBZ vertical",
        "coordinates": "elevation azimuth range",
    },
    rain_rate: {
        "units": "kg/m2/s",
        "standard_name": "rain_rate",
        "long_name": "Rain rate",
        "coordinates": "elevation azimuth range",
    },
    radar_estimated_rain_rate: {
        "units": "mm/hr",
        "standard_name": "radar_estimated_rain_rate",
        "long_name": "Radar estimated rain rate",
        "coordinates": "elevation azimuth range",
    },
    radar_echo_classification: {
        "units": "legend",
        "standard_name": "radar_echo_classification",
        "long_name": "Radar Echo classification",
        "coordinates": "elevation azimuth range",
    },
    specific_attenuation: {
        "units": "dB/km",
        "standard_name": "specific_attenuation",
        "long_name": "Specific attenuation",
        "valid_min": 0.0,
        "valid_max": 1.0,
        "coordinates": "elevation azimuth range",
    },
    specific_differential_attenuation: {
        "units": "dB/km",
        "long_name": "Specific Differential Attenuation",
        "coordinates": "elevation azimuth range",
    },
    path_integrated_attenuation: {
        "units": "dB",
        "long_name": "Path Integrated Attenuation",
        "coordinates": "elevation azimuth range",
    },
    path_integrated_differential_attenuation: {
        "units": "dB",
        "long_name": "Path Integrated Differential Attenuation",
        "coordinates": "elevation azimuth range",
    },
    # Textures
    differential_phase_texture: {
        "units": "degrees",
        "standard_name": "differential_phase_hv_texture",
        "long_name": "Texture of differential phase (PhiDP)",
        "coordinates": "elevation azimuth range",
    },
    reflectivity_texture: {
        "units": "dBZ",
        "long_name": "Texture of Reflectivity",
        "coordinates": "elevation azimuth range",
    },
    differential_reflectivity_texture: {
        "units": "dB",
        "long_name": "Texture of Differential Reflectivity",
        "coordinates": "elevation azimuth range",
    },
    cross_correlation_ratio_texture: {
        "units": "ratio",
        "long_name": "Texture of Cross Correlation Ratio",
        "coordinates": "elevation azimuth range",
    },
    # Wind retrieval fields
    eastward_wind_component: {
        "units": "meters_per_second",
        "standard_name": "eastward_wind_component",
        "long_name": "Eastward wind component",
    },
    northward_wind_component: {
        "units": "meters_per_second",
        "standard_name": "northward_wind_component",
        "long_name": "Northward wind component",
    },
    vertical_wind_component: {
        "units": "meters_per_second",
        "standard_name": "vertical_wind_component",
        "long_name": "Vertical wind component",
    },
    # profile variables
    height: {
        "units": "m",
        "long_name": "Height of radar beam",
        "standard_name": "height",
    },
    height_over_iso0: {"units": "m", "long_name": "Height of radar beam"},
    interpolated_profile: {
        "long_name": "Interpolated profile",
        "standard_name": "interpolated_profile",
        "units": "unknown",
    },
    temperature: {"units": "degC", "long_name": "Sounding Temperature"},
    clutter_filter_power_removed: {
        "units": "dB",
        "long_name": "Clutter filter power removed",
        "standard_name": "clutter_filter_power_removed",
        "coordinates": "elevation azimuth range",
    },
    # Grid metadata
    "grid_time": {
        "units": "seconds",
        "standard_name": "time",
        "long_name": "Time of grid",
        "calendar": "gregorian",
    },
    "origin_longitude": {
        "long_name": "Longitude at grid origin",
        "units": "degrees_east",
        "standard_name": "longitude",
        "valid_min": -180.0,
        "valid_max": 180.0,
    },
    "origin_latitude": {
        "long_name": "Latitude at grid origin",
        "units": "degrees_north",
        "standard_name": "latitude",
        "valid_min": -90.0,
        "valid_max": 90.0,
    },
    "origin_altitude": {
        "long_name": "Altitude at grid origin",
        "units": "m",
        "standard_name": "altitude",
    },
    "x": {
        "standard_name": "projection_x_coordinate",
        "long_name": "X distance on the projection plane from the origin",
        "axis": "X",
        "units": "m",
    },
    "y": {
        "standard_name": "projection_y_coordinate",
        "long_name": "Y distance on the projection plane from the origin",
        "axis": "Y",
        "units": "m",
    },
    "z": {
        "standard_name": "projection_z_coordinate",
        "long_name": "Z distance on the projection plane from the origin",
        "axis": "Z",
        "units": "m",
        "positive": "up",
    },
    "point_x": {
        "long_name": "Cartesian x distance of each grid point from the origin",
        "units": "meters",
    },
    "point_y": {
        "long_name": "Cartesian y distance of each grid point from the origin",
        "units": "meters",
    },
    "point_z": {
        "long_name": "Cartesian z distance of each grid point from the origin",
        "positive": "up",
        "units": "meters",
    },
    "point_longitude": {
        "long_name": "Longitude of each grid point",
        "units": "degrees_north",
    },
    "point_latitude": {
        "long_name": "Latitude of each grid point",
        "units": "degrees_east",
    },
    "point_altitude": {"long_name": "Altitude of each grid point", "units": "meters"},
    "radar_latitude": {
        "long_name": "Latitude of radars used to make the grid.",
        "units": "degrees_north",
    },
    "radar_longitude": {
        "long_name": "Longitude of radars used to make the grid.",
        "units": "degrees_east",
    },
    "radar_altitude": {
        "long_name": "Altitude of radars used to make the grid.",
        "units": "m",
    },
    "radar_time": {
        "calendar": "gregorian",
        "long_name": "Time in seconds of the volume start for each radar",
    },
    "radar_name": {
        "long_name": "Name of radar used to make the grid",
    },
}


##############################################################################
# File specific metadata
#
# These dictionaries define metadata that is to be used only when reading in
# a given type of file.  This metadata is used in place of the
# DEFAULT_METADATA when it is avialable.  The main use of these variable
# is to define field specific data, it is safe to leave some/all of these
# empty if the default metadata is acceptable.
##############################################################################

# Metadata for Sigmet/IRIS files
sigmet_metadata = {}

# Metadata for NEXRAD Level II files (Archive and CDM files)
nexrad_metadata = {
    reflectivity: {
        "units": "dBZ",
        "standard_name": "equivalent_reflectivity_factor",
        "long_name": "Reflectivity",
        "valid_max": 94.5,
        "valid_min": -32.0,
        "coordinates": "elevation azimuth range",
    },
    velocity: {
        "units": "meters_per_second",
        "standard_name": "radial_velocity_of_scatterers_away_from_instrument",
        "long_name": "Mean doppler Velocity",
        "valid_max": 95.0,
        "valid_min": -95.0,
        "coordinates": "elevation azimuth range",
    },
    spectrum_width: {
        "units": "meters_per_second",
        "standard_name": "doppler_spectrum_width",
        "long_name": "Spectrum Width",
        "valid_max": 63.0,
        "valid_min": -63.5,
        "coordinates": "elevation azimuth range",
    },
    differential_reflectivity: {
        "units": "dB",
        "standard_name": "log_differential_reflectivity_hv",
        "long_name": "log_differential_reflectivity_hv",
        "valid_max": 7.9375,
        "valid_min": -7.8750,
        "coordinates": "elevation azimuth range",
    },
    differential_phase: {
        "units": "degrees",
        "standard_name": "differential_phase_hv",
        "long_name": "differential_phase_hv",
        "valid_max": 360.0,
        "valid_min": 0.0,
        "coordinates": "elevation azimuth range",
    },
    cross_correlation_ratio: {
        "units": "ratio",
        "standard_name": "cross_correlation_ratio_hv",
        "long_name": "Cross correlation_ratio (RHOHV)",
        "valid_max": 1.0,
        "valid_min": 0.0,
        "coordinates": "elevation azimuth range",
    },
    clutter_filter_power_removed: {
        "units": "dB",
        "long_name": "Clutter filter power removed",
        "standard_name": "clutter_filter_power_removed",
        "valid_max": 73.0,
        "valid_min": 0.0,
        "coordinates": "elevation azimuth range",
    },
}

# Metadata for NEXRAD Level 3 Products
nexrad_level3_metadata = {
    radar_estimated_rain_rate: {
        "units": "inches/hour",
        "standard_name": "radar_estimated_rain_rate",
        "long_name": "Radar estimated rain rate",
        "coordinates": "elevation azimuth range",
    },
    radar_echo_classification: {
        "units": "legend",
        "standard_name": "radar_echo_classification",
        "long_name": "Radar echo classification",
        "options": (
            "0: Below Threshold (ND), "
            "10: Biological (BI), "
            "20: Anomalous Propagation/Group Clutter (GC), "
            "30: Ice Crystals (IC), "
            "40: Dry Snow (DS), "
            "50: Wet Snow (WS), "
            "60: Light and/or Moderate Rain (RA), "
            "70: Heavy Rain (HR), "
            "80: Big Drops (rain) (BD), "
            "90: Graupel (GR), "
            "100: Hail, possibly with rain (HA), "
            "110: Large Hail (LH), "
            "120: Giant Hail (GH), "
            "140: Unknown Classification (UK), "
            "150: Range Folded (RH)"
        ),
        "coordinates": "elevation azimuth range",
    },
}

# Metadata for CF/Radial files
cfradial_metadata = {}

# Metadata for MDV files
mdv_metadata = {}

# Metadata for RSL files
rsl_metadata = {}

# Metadata for CSU-CHILL, CHL files
chl_metadata = {}

FILE_SPECIFIC_METADATA = {  # Required
    "sigmet": sigmet_metadata,
    "nexrad_archive": nexrad_metadata,
    "nexrad_cdm": nexrad_metadata,
    "nexrad_level3": nexrad_level3_metadata,
    "cfradial": cfradial_metadata,
    "mdv": mdv_metadata,
    "rsl": rsl_metadata,
    "chl": chl_metadata,
}

##############################################################################
# Field name mapping
#
# These dictionaries map file field names or data types to a radar field
# name.  These are used to populate the radar.fields dictionary during a read
# in Py-ART.  A value of None will not include that field in the radar object.
# These can be over-ridden on a per-read basis using the field_mapping
# parameter, or using setting the file_field_names parameter to True.
##############################################################################

# Sigmet/IRIS file field mapping
# Note that multiple sigmet fields map to the same radar field, if
# more than one of these fields are present the radar field will be
# overwritten with the last sigmet field.
sigmet_field_mapping = {
    # Sigmet data type :field name              # (Data_type) Description
    "XHDR": None,  # (0) Extended Header
    "DBT": total_power,  # (1) Total Power
    "DBZ": reflectivity,  # (2) Reflectivity
    "VEL": velocity,  # (3) Velocity
    "WIDTH": spectrum_width,  # (4) Width
    "ZDR": differential_reflectivity,  # (5) Diff. reflectivity
    "DBZC": corrected_reflectivity,  # (7) Corrected reflectivity
    "DBT2": total_power,  # (8) Total Power
    "DBZ2": reflectivity,  # (9) Reflectivity
    "VEL2": velocity,  # (10) Velocity
    "WIDTH2": spectrum_width,  # (11) Width
    "ZDR2": differential_reflectivity,  # (12) Diff. reflectivity
    "RAINRATE2": radar_estimated_rain_rate,  # (13) Rainfall rate
    "KDP": specific_differential_phase,  # (14) KDP (diff. phase)
    "KDP2": specific_differential_phase,  # (15) KDP (diff. phase)
    "PHIDP": differential_phase,  # (16) PhiDP (diff. phase)
    "VELC": corrected_velocity,  # (17) Corrected velocity
    "SQI": normalized_coherent_power,  # (18) SQI
    "RHOHV": cross_correlation_ratio,  # (19) RhoHV
    "RHOHV2": cross_correlation_ratio,  # (20) RhoHV
    "DBZC2": corrected_reflectivity,  # (21) Corrected Reflec.
    "VELC2": corrected_velocity,  # (21) Corrected Velocity
    "SQI2": normalized_coherent_power,  # (23) SQI
    "PHIDP2": differential_phase,  # (24) PhiDP (diff. phase)
    "LDRH": linear_depolarization_ratio_h,  # (25) LDR xmt H, rcv V
    "LDRH2": linear_depolarization_ratio_h,  # (26) LDR xmt H, rcv V
    "LDRV": linear_depolarization_ratio_v,  # (27) LDR xmt V, rcv H
    "LDRV2": linear_depolarization_ratio_v,  # (28) LDR xmt V, rcv H
    "HEIGHT": None,  # (32) Height (1/10 km)
    "VIL2": None,  # (33) Linear Liquid
    "RAW": None,  # (34) Raw Data
    "SHEAR": None,  # (35) Wind Shear
    "DIVERGE2": None,  # (36) Divergence
    "FLIQUID2": None,  # (37) Floated liquid
    "USER": None,  # (38) User type
    "OTHER": None,  # (39) Unspecified
    "DEFORM2": None,  # (40) Deformation
    "VVEL2": None,  # (41) Vertical velocity
    "HVEL2": None,  # (42) Horizontal velocity
    "HDIR2": None,  # (43) Horiz. wind direction
    "AXDIL2": None,  # (44) Axis of dilation
    "TIME2": None,  # (45) Time in seconds
    "RHOH": None,  # (46) Rho, xmt H, rcv V
    "RHOH2": None,  # (47) Rho, xmt H, rcv V
    "RHOV": None,  # (48) Rho, xmt V, rcv H
    "RHOV2": None,  # (49) Rho, xmt V, rcv H
    "PHIH": None,  # (50) Phi, xmt H, rcv V
    "PHIH2": None,  # (51) Phi, xmt H, rcv V
    "PHIV": None,  # (52) Phi, xmt V, rcv H
    "PHIV2": None,  # (53) Phi, xmt V, rcv H
    "USER2": None,  # (54) User type
    "HCLASS": radar_echo_classification,  # (55) Hydrometeor class
    "HCLASS2": radar_echo_classification,  # (56) Hydrometeor class
    "ZDRC": corrected_differential_reflectivity,
    # (57) Corrected diff. refl.
    "ZDRC2": corrected_differential_reflectivity,
    # (58) Corrected diff. refl.
    "UNKNOWN_59": None,  # Unknown field
    "UNKNOWN_60": None,  # Unknown field
    "UNKNOWN_61": None,  # Unknown field
    "UNKNOWN_62": None,  # Unknown field
    "UNKNOWN_63": None,  # Unknown field
    "UNKNOWN_64": None,  # Unknown field
    "UNKNOWN_65": None,  # Unknown field
    "UNKNOWN_66": None,  # Unknown field
    "PMI8": None,  # (75) Polarimetric Meteo Index
    "PMI16": None,  # (76) Polarimetric Meteo Index
    "LOG8": None,  # (77) Log receiver SNR
    "LOG16": None,  # (78) Log receiver SNR
    "CSP8": None,  # (77) Doppler channel clutter power
    "CSP16": None,  # (78) Doppler channel clutter power
    # there may be more field, add as needed
}


# NEXRAD Level II Archive files
nexrad_archive_field_mapping = {
    # NEXRAD field: radar field name
    "REF": reflectivity,
    "VEL": velocity,
    "SW": spectrum_width,
    "ZDR": differential_reflectivity,
    "PHI": differential_phase,
    "RHO": cross_correlation_ratio,
    "CFP": clutter_filter_power_removed,
}

# NEXRAD Level II CDM files
nexrad_cdm_field_mapping = {
    # CDM variable name (without _HI): radar field name
    "Reflectivity": reflectivity,
    "RadialVelocity": velocity,
    "SpectrumWidth": spectrum_width,
    "DifferentialReflectivity": differential_reflectivity,
    "DifferentialPhase": differential_phase,
    "CorrelationCoefficient": cross_correlation_ratio,
}

# NEXRAD Level 3 Product files.
nexrad_level3_mapping = {
    # Message code : field name         # Product name
    19: reflectivity,  # Base Reflectivity
    20: reflectivity,  # Base Reflectivity
    25: velocity,  # Base Velocity
    27: velocity,  # Base Velocity
    28: spectrum_width,  # Base Spectrum Width
    30: spectrum_width,  # Base Spectrum Width
    32: reflectivity,  # Digital Hybrid Scan Reflectivity
    34: None,  # Clutter Filter Control
    56: velocity,  # Storm Relative Mean Radial Velocity
    78: radar_estimated_rain_rate,  # Surface Rainfall Accum. (1 hr)
    79: radar_estimated_rain_rate,  # Surface Rainfall Accum. (3 hr)
    80: radar_estimated_rain_rate,  # Storm Total Rainfall Accumulation
    94: reflectivity,  # Base Reflectivity Data Array
    99: velocity,  # Base Velocity Data Array
    134: None,  # High Resolution VIL
    135: None,  # Enhanced Echo Tops
    138: radar_estimated_rain_rate,  # Digital Storm Total Precipitation
    159: differential_reflectivity,  # Digital Differential Reflectivity
    161: cross_correlation_ratio,  # Digital Correlation Coefficient
    163: specific_differential_phase,  # Digital Specific Differential Phase
    165: radar_echo_classification,  # Digital Hydrometeor Classification
    169: radar_estimated_rain_rate,  # One Hour Accumulation
    170: radar_estimated_rain_rate,  # Digital Accumulation Array
    171: radar_estimated_rain_rate,  # Storm Total Accumulation
    172: radar_estimated_rain_rate,  # Digital Storm Total Accumulation
    173: radar_estimated_rain_rate,  # Digital User-Selectable Accum.
    174: radar_estimated_rain_rate,  # Digital 1 hr Diff. Accum.
    175: radar_estimated_rain_rate,  # Digital Storm Total Diff. Accum.
    176: radar_estimated_rain_rate,  # Digital Inst. Precipitation Rate
    177: radar_echo_classification,  # Hybrid Hydrometeor Classification
    181: reflectivity,  # Base Reflectivity
    182: velocity,  # Base Velocity
    186: reflectivity,  # Base Reflectivity
}

# MDV files
mdv_field_mapping = {
    # MDV moment: radar field name
    "DBZ_F": reflectivity,
    "VEL_F": velocity,
    "WIDTH_F": spectrum_width,
    "ZDR_F": differential_reflectivity,
    "RHOHV_F": cross_correlation_ratio,
    "NCP_F": normalized_coherent_power,
    "KDP_F": specific_differential_phase,
    "PHIDP_F": differential_phase,
    "VEL_COR": corrected_velocity,
    "PHIDP_UNF": unfolded_differential_phase,
    "KDP_SOB": corrected_specific_differential_phase,
    "DBZ_AC": corrected_reflectivity,
    # repeated integer moments
    "DBZ": reflectivity,
    "VEL": velocity,
    "WIDTH": spectrum_width,
    "ZDR": differential_reflectivity,
    "RHOHV": cross_correlation_ratio,
    "NCP": normalized_coherent_power,
    "KDP": specific_differential_phase,
    "PHIDP": differential_phase,
}

# CF/Radial files
cfradial_field_mapping = {}

# RSL files
# Note that multiple RSL field map to the same radar field, if
# more than one of these fields are present in the RSL data structure
# the radar field will be overwritten with the last field.
rsl_field_mapping = {
    # RSL 2 letter field: radar field           # RSL description
    "DZ": reflectivity,  # reflectivity
    "VR": velocity,  # velocity
    "SW": spectrum_width,  # spectrum width
    "CZ": corrected_reflectivity,  # corrected reflectivity
    "ZT": reflectivity,  # uncorrected reflectivity
    "DR": differential_reflectivity,  # differential reflectivity
    "LR": differential_reflectivity,  # another diff. reflectivity
    "ZD": differential_reflectivity,  # another diff. reflectivity
    "DM": None,  # received power
    "RH": cross_correlation_ratio,  # RhoHV
    "PH": differential_phase,  # PhiDP
    "XZ": None,  # X-band reflectivity
    "CD": corrected_differential_reflectivity,  # Corrected DR.
    "MZ": None,  # DZ mask
    "MD": None,  # DR Mask
    "ZE": corrected_reflectivity,  # edited reflectivity
    "VE": corrected_velocity,  # edited velocity
    "KD": specific_differential_phase,  # specific diff. phase
    "TI": None,  # TIME (unknown)
    "DX": None,  # ???
    "CH": None,  # ???
    "AH": None,  # ???
    "CV": None,  # ???
    "AV": None,  # ???
    "SQ": normalized_coherent_power,  # Signal Quality Index (sigmet)
    "VS": None,  # Radial Vel. combined
    "VL": None,  # Radial Vel. combined
    "VG": None,  # Radial Vel. combined
    "VT": None,  # Radial Vel. combined
    "NP": normalized_coherent_power,  # Normalized Coherent Power
    "HC": radar_echo_classification,  # Hydroclass
    "VC": None,  # Radial Vel. Corrected.
    "V2": None,  # Radial Vel cut 2
    "S2": None,  # Spectrum width cut 2
    "V3": None,  # Radial Vel cut 3
    "S3": None,  # Spectrum width cut 3
}

chl_field_mapping = {
    # Chill field name : radar field name
    "Z": reflectivity,
    "V": velocity,
    "W": spectrum_width,
    "ZDR": differential_reflectivity,
    "LDRH": linear_depolarization_ratio_h,
    "LDRV": linear_depolarization_ratio_v,
    b"\xce\xa8 DP".decode("utf-8"): differential_phase,
    "KDP": specific_differential_phase,
    b"\xcf\x81 HV".decode("utf-8"): cross_correlation_ratio,
    "NCP": normalized_coherent_power,
    # These fields are not mapped by default
    "H Re(lag 1)": None,  # Real part of lag-1 correlation, H Channel
    "V Re(lag 2)": None,  # Real part of lag-2 correlation, V Channel
    "VAvgQ": None,  # Average Q, V Channel
    "V Im(lag 1)": None,  # Imaginary part of lag-1 correlation, V Channel
    "HAvgQ": None,  # Average Q, H Channel
    "H Im(lag 2)": None,  # Imaginary part of lag-2 correlation, H Channel
    "V lag 0": None,  # Absolute value of lag-0 correlation, V Channel
    "H lag 0": None,  # Absolute value of lag-0 correlation, H Channel
    "H lag 0 cx": None,  # Absolute value of lag-0 cross correlation,
    # H Channel
    "H Im(lag 1)": None,  # Imaginary part of lag-1 correlation, H Channel
    "H Re(lag 2)": None,  # Real part of lag-2 correlation, H Channel
    "V lag 0 cx": None,  # Absolute value of lag-0 cross correlation,
    # V Channel
    "V Re(lag 1)": None,  # Real part of lag-1 correlation, V Channel
    "V Im(lag 2)": None,  # Imaginary part of lag-2 correlation, V Channel
    "HV lag 0 I": None,  # Real part of cross channel correlation at lag-0
    "HV lag 0 Q": None,  # Imaginary part of cross channel correlation at
    # lag-0
    "VAvgI": None,  # Average I, V Channel
    "HAvgI": None,  # Average I, H Channel
    b"\xcf\x81 HCX".decode("utf-8"): None,  # H Co to Cross Correlation
    b"\xcf\x81 VCX".decode("utf-8"): None,  # V Co to Cross Correlation
}

# GAMIC HDF5 files
gamic_field_mapping = {
    # Description of GAMIC fields taken from Radx source code file:
    # GamicHdf5RadxFile.cc
    # General notes on field names
    # ----------------------------
    # * An 'U' prefix indicated the moment was corrected from a timeseries
    #   that was not clutter corrected.  Fields without a 'U' prefix are
    #   calculated from a timeseries that has been clutter corrected.
    # * An 'A' prefix indicates that the moment has been corrected
    #   for rainfall attenuation.
    # * 'h' and 'v' endings indicate a the horizontal and vertical channels
    # GAMIC field name: radar field name
    "I": None,  # In phase signal
    "Q": None,  # Quadrature signal
    "Z": corrected_reflectivity,
    # Reflectivity, corrected for 2nd trip & clutter
    "UZ": reflectivity,
    # Uncorrected reflectivity
    "Zh": corrected_reflectivity,
    "ZH": corrected_reflectivity,
    # Corrected reflectivity, horizontal channel
    "Zv": None,
    "ZV": None,  # Corrected reflectivity, vertical channel
    "UZh": reflectivity,
    "UH": reflectivity,
    # Uncorrected reflectivity, horizontal channel
    "UZv": None,
    "UV": None,  # Uncorrected reflectivity, vertical channel
    "AZh": None,  # Refl., rainfall atten. & clutter corrected, h chan.
    # 'F' in velocity fields indicate folded velocities (no de-aliasing),
    # velocities fields without an F have been unfolded.
    "V": corrected_velocity,
    # Unfolded velocity from corrected timeseries
    "VF": velocity,  # Folded velocity from corr. t.s.
    "UV": None,  # Unfolded velocity from uncorrected timeseries
    "UVF": None,  # Folded velcity from uncorr. t.s.
    "Vh": corrected_velocity,
    "VH": corrected_velocity,
    # Velocity from corr. t.s., horizontal channel
    "Vv": None,
    "VV": None,  # Velocity from corr. t.s., vertical channel
    "UVh": None,  # Velocity from uncorr. t.s., horizontal channel
    "UVv": None,  # Velocity from uncorr. t.s., vertical channel
    "VFh": velocity,
    # Folded velocity, corr. t.s., horizontal channel
    "VFv": None,  # Folded velocity, corr. t.s., vertical channel
    "UnV": None,  # Folded velocity from uncorrected timeseries
    "UnVFh": None,  # Folded velocity, uncorr. t.s., horizontal channel
    "UnVFv": None,  # Folded velocity, uncorr. t.s., vertical channel
    # 'C' in spectral width fields indicate that the field has been
    # corrected for decorrelation causes by antenna rotation
    "W": spectrum_width,
    # Spectral width from clutter corrected time series.
    "UW": None,  # Spectral width from uncorrected timeseries.
    "CW": None,  # Spec. width, antenna rotation corrected, corr. t.s.
    "UCW": None,  # Spec. width, antenna rotation corrected, uncorr t.s.
    "Wh": spectrum_width,
    "WH": spectrum_width,
    # Spectral width, corr t.s., horizontal channel
    "Wv": None,
    "WV": None,  # Spectral width, corr t.s., vertical channel
    "UWh": None,  # Spectral width, uncorr t.s., horizontal channel
    "UWv": None,  # Spectral width, uncorr t.s., vertical channel
    "CWh": None,  # Spec. width, antenna rot. corr., horizontal channel
    "CWv": None,  # Spec. width, antenna rot. corr., vertical channel
    "SQI": normalized_coherent_power,
    # Signal quality index
    "SQIh": normalized_coherent_power,
    # Signal quality index, horizontal channel
    "SQIv": None,  # Signal quality index, vertical channel
    "CCOR": None,  # Clutter power correction
    "CCORh": None,  # Clutter power correction, horizontal channel
    "CCORv": None,  # Clutter power correction, vertical channel
    "SIGPOW": None,  # Singal Power
    "SNR": None,  # Raw signal to noise ratio
    "SNRh": None,  # Raw signal to noise ratio, horizontal channel
    "SNRv": None,  # Raw signal to noise ration, vertical channel
    "DFT": None,  # Signal spectrum amplitude
    "DFTh": None,  # Signal spectrum amplitude, horizontal channel
    "DFTv": None,  # Signal spectrum amplitude, vertical channel
    "LOG": None,  # Logarithmic amplitude 10*log Isq + Qsq
    "LOGh": None,  # Logarithmic amplitude, horizontal channel
    "LOGv": None,  # Logarithmic amplitude, vertical channel
    "CMAP": None,  # Censor map
    # A '1' ending on differential reflectivity fields indicates that the
    # moment has calculated using a 1st LAG algorithm.
    "ZDR": corrected_differential_reflectivity,
    # Differential reflectivity from corrected timeseries
    "UZDR": differential_reflectivity,
    # Diff. refl. from uncorrected timeseries
    "AZDR": None,  # Diff. refl., rainfall atten. corr., corr t.s.
    "ZDR1": None,  # Diff. refl., corr. t.s., 1st LAG algo.
    "UZDR1": None,  # Diff. refl., uncorr. t.s., 1st LAG algo.
    "AZDR1": None,  # Diff. refl., rain. atten. corr., corr. t.s., 1st LAG
    "PHI": corrected_differential_phase,
    # Differential phase from corrected timeseries
    "PHIDP": corrected_differential_phase,
    # Differential phase from corrected timeseries
    "UPHIDP": differential_phase,
    # Diff. phase from uncorrected timeseries.
    "PHIH": None,  # Diff. phase, corr. t.s., horizontal channel
    "UPHIH": None,  # Diff. phase, uncorr t.s., hortizontal channel
    "KDP": specific_differential_phase,
    # Specific differential phase
    "RHO": cross_correlation_ratio,
    # Cross correlation coefficient from corrected t.s.
    "RHOHV": cross_correlation_ratio,
    # Cross correlation coefficient from corrected t.s.
    "URHOHV": None,  # Cross correlation coefficient from uncorr. t.s.
    "RHOH": None,  # Cross corr., corr. t.s., horizontal transmit only
    "URHOH": None,  # Cross corr., uncorr t.s., horizontal transmit only
    "LDR": linear_depolarization_ratio,
    # Linear depolarization ratio from corr. t.s.
    "ULDR": None,  # Linear depolarization ratio  from uncorr. t.s.
}

# UF field mappings
uf_field_mapping = {
    # UF 2 letter field: radar field
    "DZ": reflectivity,
    "CZ": corrected_reflectivity,
    "ZE": corrected_reflectivity,
    "ZT": total_power,
    "UZ": total_power,
    "VR": velocity,
    "VE": corrected_velocity,
    "SW": spectrum_width,
    "ZD": differential_reflectivity,
    "DR": corrected_differential_reflectivity,
    "CD": corrected_differential_reflectivity,
    "LR": linear_depolarization_ratio,
    "PH": differential_phase,
    "KD": specific_differential_phase,
    "RH": cross_correlation_ratio,
    "SQ": normalized_coherent_power,
    "NP": normalized_coherent_power,
    "HC": radar_echo_classification,
}

# Mapping used when writing UF files
write_uf_mapping = {
    # UF 2 letter field: radar field
    reflectivity: "DZ",
    corrected_reflectivity: "CZ",
    total_power: "ZT",
    velocity: "VR",
    corrected_velocity: "VE",
    spectrum_width: "SW",
    differential_reflectivity: "ZD",
    corrected_differential_reflectivity: "DR",
    linear_depolarization_ratio: "LR",
    differential_phase: "PH",
    specific_differential_phase: "KD",
    cross_correlation_ratio: "RH",
    normalized_coherent_power: "SQ",
    radar_echo_classification: "HC",
}

FIELD_MAPPINGS = {  # Required variable
    "sigmet": sigmet_field_mapping,
    "nexrad_archive": nexrad_archive_field_mapping,
    "nexrad_cdm": nexrad_cdm_field_mapping,
    "nexrad_level3": nexrad_level3_mapping,
    "cfradial": cfradial_field_mapping,
    "mdv": mdv_field_mapping,
    "rsl": rsl_field_mapping,
    "chl": chl_field_mapping,
    "gamic": gamic_field_mapping,
    "uf": uf_field_mapping,
    "write_uf": write_uf_mapping,
}


def velocity_limit(container=None, selection=0):
    import pyart

    if isinstance(container, pyart.core.Radar):
        try:
            if selection >= 0 and selection < container.nsweeps:
                vel = container.get_nyquist_vel(selection, check_uniform=False)
            else:
                vel = container.get_nyquist_vel(0, check_uniform=False)
            return (-vel, vel)
        except LookupError:
            return (-30.0, 30.0)
    else:
        return (-30.0, 30.0)


def spectrum_width_limit(container=None, selection=0):
    import pyart

    if isinstance(container, pyart.core.Radar):
        try:
            if selection >= 0 and selection < container.nsweeps:
                vel = container.get_nyquist_vel(selection, check_uniform=False)
            else:
                vel = container.get_nyquist_vel(0, check_uniform=False)
            return (0, vel)
        except LookupError:
            return (0, 30.0)
    else:
        return (0, 30.0)


# map each field to a colormap

DEFAULT_FIELD_COLORMAP = {
    # field name : colormap
    reflectivity: "pyart_HomeyerRainbow",
    corrected_reflectivity: "pyart_HomeyerRainbow",
    total_power: "pyart_HomeyerRainbow",
    signal_to_noise_ratio: "pyart_Carbone17",
    velocity: "pyart_BuDRd18",
    corrected_velocity: "pyart_BuDRd18",
    simulated_velocity: "pyart_BuDRd18",
    eastward_wind_component: "pyart_BuDRd18",
    northward_wind_component: "pyart_BuDRd18",
    vertical_wind_component: "pyart_BuDRd18",
    spectrum_width: "pyart_NWS_SPW",
    normalized_coherent_power: "pyart_Carbone17",
    differential_reflectivity: "pyart_RefDiff",
    corrected_differential_reflectivity: "pyart_RefDiff",
    clutter_filter_power_removed: "pyart_RefDiff",
    cross_correlation_ratio: "pyart_RefDiff",
    logarithmic_cross_correlation_ratio: "pyart_RefDiff",
    differential_phase: "pyart_Wild25",
    unfolded_differential_phase: "pyart_Wild25",
    corrected_differential_phase: "pyart_Wild25",
    specific_differential_phase: "pyart_Theodore16",
    corrected_specific_differential_phase: "pyart_Theodore16",
    linear_depolarization_ratio: "pyart_SCook18",
    linear_depolarization_ratio_h: "pyart_SCook18",
    linear_depolarization_ratio_v: "pyart_SCook18",
    circular_depolarization_ratio: "pyart_SCook18",
    rain_rate: "pyart_RRate11",
    radar_estimated_rain_rate: "pyart_RRate11",
    radar_echo_classification: "pyart_LangRainbow12",
    specific_attenuation: "pyart_Carbone17",
    differential_phase_texture: "pyart_BlueBrown11",
    height: "pyart_SCook18",
    interpolated_profile: "pyart_SCook18",
    noisedBZ_hh: "pyart_HomeyerRainbow",
    noisedBZ_vv: "pyart_HomeyerRainbow",
    # Additional reflectivity like fields
    "CZ": "pyart_HomeyerRainbow",
    "DZ": "pyart_HomeyerRainbow",
    "AZ": "pyart_HomeyerRainbow",
    "Z": "pyart_HomeyerRainbow",
    "dbz": "pyart_HomeyerRainbow",
    "DBZ": "pyart_HomeyerRainbow",
    "dBZ": "pyart_HomeyerRainbow",
    "DBZH": "pyart_HomeyerRainbow",
    "DBZ_S": "pyart_HomeyerRainbow",
    "DBZ_K": "pyart_HomeyerRainbow",
    "reflectivity_horizontal": "pyart_HomeyerRainbow",
    "corr_reflectivity": "pyart_HomeyerRainbow",
}

# map each field to a limit or a limit function

DEFAULT_FIELD_LIMITS = {
    # field name : limits
    reflectivity: (-30.0, 75.0),
    corrected_reflectivity: (-30.0, 75.0),
    total_power: (-30.0, 75.0),
    signal_to_noise_ratio: (-20, 30.0),
    velocity: velocity_limit,
    corrected_velocity: velocity_limit,
    eastward_wind_component: velocity_limit,
    northward_wind_component: velocity_limit,
    vertical_wind_component: velocity_limit,
    spectrum_width: spectrum_width_limit,
    normalized_coherent_power: (0.0, 1.0),
    differential_reflectivity: (-1.0, 8.0),
    corrected_differential_reflectivity: (-1.0, 8.0),
    cross_correlation_ratio: (0.5, 1.05),
    differential_phase: (-180, 180.0),
    unfolded_differential_phase: (-360, 360.0),
    corrected_differential_phase: (-360, 360.0),
    specific_differential_phase: (-2.0, 5.0),
    corrected_specific_differential_phase: (-2.0, 5.0),
    linear_depolarization_ratio: (-40.0, 0.0),
    linear_depolarization_ratio_h: (-40.0, 0.0),
    linear_depolarization_ratio_v: (-40.0, 0.0),
    rain_rate: (0.0, 50.0),
    radar_estimated_rain_rate: (0.0, 50.0),
    radar_echo_classification: (0, 11),
    specific_attenuation: (0.0, 10.0),
    differential_phase_texture: (0, 180.0),
    height: (0, 20000),
    interpolated_profile: (0, 10000),
    # Additional reflectivity like fields
    "CZ": (-10.0, 65.0),
    "DZ": (-10.0, 65.0),
    "AZ": (-10.0, 65.0),
    "Z": (-10.0, 65.0),
    "dbz": (-10.0, 65.0),
    "DBZ": (-10.0, 65.0),
    "dBZ": (-10.0, 65.0),
    "DBZH": (-10.0, 65.0),
    "DBZ_S": (-10.0, 65.0),
    "DBZ_K": (-10.0, 65.0),
    "reflectivity_horizontal": (-10.0, 65.0),
    "corr_reflectivity": (-10.0, 65.0),
}



"""
=========================================
Calculate and Plot hydrometeor classification
=========================================

Calculates a hydrometeor classification and displays the results
"""

# Author: Daniel Wolfensberger (daniel.wolfensberger@meteoswiss.ch)
# License: BSD 3 clause

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from open_radar_data import DATASETS

import pyart

# Read in a sample file
filename = DATASETS.fetch("MLL2217907250U.003.nc")
radar = pyart.io.read_cfradial(filename)

# Read temperature preinterpolated from NWP model
filename = DATASETS.fetch("20220628072500_savevol_COSMO_LOOKUP_TEMP.nc")
nwp_temp = pyart.io.read_cfradial(filename)

# Add temperature to radar object as new field
radar.add_field("temperature", nwp_temp.fields["temperature"])

# Compute attenuation
out = pyart.correct.calculate_attenuation_zphi(
    radar,
    phidp_field="uncorrected_differential_phase",
    temp_field="temperature",
    temp_ref="temperature",
)
spec_at, pia, cor_z, spec_diff_at, pida, cor_zdr = out
radar.add_field("corrected_reflectivity", cor_z)
radar.add_field("corrected_differential_reflectivity", cor_zdr)
radar.add_field("specific_attenuation", spec_at)

# Compute KDP
kdp, _, _ = pyart.retrieve.kdp_maesaka(
    radar, psidp_field="uncorrected_differential_phase"
)
radar.add_field("specific_differential_phase", kdp)

# Compute hydrometeor classification
hydro = pyart.retrieve.hydroclass_semisupervised(
    radar,
    refl_field="corrected_reflectivity",
    zdr_field="corrected_differential_reflectivity",
    kdp_field="specific_differential_phase",
    rhv_field="uncorrected_cross_correlation_ratio",
    temp_field="temperature",
)

radar.add_field("radar_echo_classification", hydro)

# Display hydrometeor classification with categorical colormap
fig, ax = plt.subplots(1, 1, figsize=(6, 6))
display = pyart.graph.RadarDisplay(radar)

labels = ["NC", "AG", "CR", "LR", "RP", "RN", "VI", "WS", "MH", "IH/HDG"]
ticks = np.arange(len(labels))
boundaries = np.arange(-0.5, len(labels))
norm = mpl.colors.BoundaryNorm(boundaries, 256)

cax = display.plot_ppi(
    "radar_echo_classification", 0, ax=ax, norm=norm, ticks=ticks, ticklabs=labels
)

ax.set_xlim([-50, 50])
ax.set_ylim([-50, 50])
ax.set_aspect("equal", "box")

# For info
# NC = not classified
# AG = aggregates
# CR = ice crystals
# LR = light rain
# RP = rimed particles
# RN = rain
# VI = vertically oriented ice
# WS = wet snow
# MH = melting hail
# IH/HDG = dry hail / high density graupel
